#!/usr/bin/env python3
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

import argparse
import pathlib
import sys
import unittest

from modules.test_utils import AbstractBlenderRunnerTest


class AbstractFFmpegTest(AbstractBlenderRunnerTest):
    @classmethod
    def setUpClass(cls):
        cls.blender = args.blender
        cls.testdir = pathlib.Path(args.testdir)


class AbstractFFmpegSequencerTest(AbstractFFmpegTest):
    def get_script_for_file(self, filename: pathlib.Path) -> str:
        movie = self.testdir / filename
        return \
            "import bpy; " \
            "bpy.context.scene.sequence_editor_create(); " \
            "strip = bpy.context.scene.sequence_editor.sequences.new_movie(" \
            "'test_movie', %r, channel=1, frame_start=1); " \
            "print(f'fps:{strip.fps}'); " \
            "print(f'duration:{strip.frame_final_duration}'); " % movie.as_posix()

    def get_movie_file_field(self, filename: pathlib.Path, field: str) -> str:
        script = self.get_script_for_file(filename)
        output = self.run_blender('', script)
        prefix = field + ":"
        for line in output.splitlines():
            if line.startswith(prefix):
                return line.split(':')[1]
        return ""

    def get_movie_file_field_float(self, filename: pathlib.Path, field: str) -> float:
        return float(self.get_movie_file_field(filename, field))

    def get_movie_file_field_int(self, filename: pathlib.Path, field: str) -> float:
        return int(self.get_movie_file_field(filename, field))

    def get_movie_file_fps(self, filename: pathlib.Path) -> float:
        return self.get_movie_file_field_float(filename, "fps")

    def get_movie_file_duration(self, filename: pathlib.Path) -> float:
        return self.get_movie_file_field_int(filename, "duration")


class FPSDetectionTest(AbstractFFmpegSequencerTest):
    def test_T51153(self):
        self.assertAlmostEqual(
            self.get_movie_file_fps('T51153_bad_clip_2.mts'),
            29.97,
            places=2)

    def test_T53857(self):
        self.assertAlmostEqual(
            self.get_movie_file_fps('T53857_2018-01-22_15-30-49.mkv'),
            30.0,
            places=2)

    def test_T54148(self):
        self.assertAlmostEqual(
            self.get_movie_file_fps('T54148_magn_0.mkv'),
            1.0,
            places=2)

    def test_T68091(self):
        self.assertAlmostEqual(
            self.get_movie_file_fps('T68091-invalid-nb_frames-at-10fps.mp4'),
            10.0,
            places=2)
        self.assertEqual(
            self.get_movie_file_duration('T68091-invalid-nb_frames-at-10fps.mp4'),
            10)

    def test_T54834(self):
        self.assertEqual(
            self.get_movie_file_duration('T54834.ogg'),
            50)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--blender', required=True)
    parser.add_argument('--testdir', required=True)
    args, remaining = parser.parse_known_args()

    unittest.main(argv=sys.argv[0:1] + remaining)



"""
Created on Sep 2, 2011

@author: vencax
"""

from .model import DxfDrawing

try:
    from . import dxfLibrary as DXF
    #reload(DXF)
    #reload(dxfLibrary)
    #from dxfLibrary import *
except Exception:
    raise Exception("No dxfLibrary.py module in Blender script folder found!")

#------------------------------------------------------
#def col2RGB(color):
#    return [int(floor(255*color[0])),
#            int(floor(255*color[1])),
#            int(floor(255*color[2]))]
#
#global dxfColors
#dxfColors=None
#
##------------------------------------------------------
#def col2DXF(rgbcolor):
#    global dxfColors
#    if dxfColors is None:
#        from dxfColorMap import color_map
#        dxfColors = [(tuple(color),idx) for idx, color in color_map.iteritems()]
#        dxfColors.sort()
#    entry = (tuple(rgbcolor), -1)
#    dxfColors.append(entry)
#    dxfColors.sort()
#    i = dxfColors.index(entry)
#    dxfColors.pop(i)
#    return dxfColors[i-1][1]

"""
v = [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,\
28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,46,47,58,59,60,61,62,63,64,91,92,93,94,96,\
123,124,125,126,127,128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,146,147,148,149,150,\
151,152,153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,\
171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,\
191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,\
211,212,213,214,215,216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,\
231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254]
invalid = ''.join([chr(i) for i in v])
del v, i
"""
#TODO: validDXFr14 = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.'
validDXFr12 = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_'
NAMELENGTH_MAX = 80   #max_obnamelength in DXF, (limited to 256? )
#------------------------------------------------------
def cleanName(name, valid):
    validname = ''
    for ch in name:
        if ch not in valid:    ch = '_'
        validname += ch
    return validname

#------------------------------------------------------
def validDXFr12name(str_name):
    dxfname = str(str_name)
    dxfname = dxfname[:NAMELENGTH_MAX].upper()
    dxfname = cleanName(dxfname, validDXFr12)
    return dxfname

class MigiusDXFLibDrawing(DxfDrawing):
    """ Drawing that can convert itself into MIGIUS DXFLib stuff objects """

    def convert(self, file, **kwargs):
        drawing = self._write()
        for type, ents in self._entities.items():
            self._processEntityArray(drawing, type, ents)
        for v in self._views:
            drawing.views.append(DXF.View(**v))
        for v in self._vports:
            drawing.vports.append(DXF.VPort(**v))
#        for l in self._layers:
        drawing.saveas(file)

    def _write(self):
        # init Drawing ---------------------
        d=DXF.Drawing()
        # add Tables -----------------
        # initialized automatic: d.blocks.append(b)                 #section BLOCKS
        # initialized automatic: d.styles.append(DXF.Style())            #table STYLE

        #table LTYPE ---------------
        #d.linetypes.append(DXF.LineType(name='CONTINUOUS',description='--------',elements=[0.0]))
        d.linetypes.append(DXF.LineType(name='DOT',description='. . . . . . .',elements=[0.25, 0.0, -0.25]))
        d.linetypes.append(DXF.LineType(name='DASHED',description='__ __ __ __ __',elements=[0.8, 0.5, -0.3]))
        d.linetypes.append(DXF.LineType(name='DASHDOT',description='__ . __ . __ .',elements=[1.0, 0.5, -0.25, 0.0, -0.25]))
        d.linetypes.append(DXF.LineType(name='DIVIDE',description='____ . . ____ . . ',elements=[1.25, 0.5, -0.25, 0.0, -0.25, 0.0, -0.25]))
        d.linetypes.append(DXF.LineType(name='BORDER',description='__ __ . __ __ . ',elements=[1.75, 0.5, -0.25, 0.5, -0.25, 0.0, -0.25]))
        d.linetypes.append(DXF.LineType(name='HIDDEN',description='__ __ __ __ __',elements=[0.4, 0.25, -0.25]))
        d.linetypes.append(DXF.LineType(name='CENTER',description='____ _ ____ _ __',elements=[2.0, 1.25, -0.25, 0.25, -0.25]))

        #d.vports.append(DXF.VPort('*ACTIVE'))
        d.vports.append(DXF.VPort('*ACTIVE',center=(-5.0,1.0),height=10.0))
        #d.vports.append(DXF.VPort('*ACTIVE',leftBottom=(-100.0,-60.0),rightTop=(100.0,60.0)))
        #d.views.append(DXF.View('Normal'))      #table view
        d.views.append(DXF.ViewByWindow('BF_TOPVIEW',leftBottom=(-100,-60),rightTop=(100,60)))  #idem

        return d

    def _processEntityArray(self, drawing, type, ents):
        if type=='Point':
            for e in ents:
                drawing.append(DXF.Point(**e))
        elif type=='Line':
            for e in ents:
                drawing.append(DXF.Line(**e))
        elif type=='PolyLine':
            for e in ents:
                drawing.append(DXF.PolyLine(**e))
        elif type=='Face':
            for e in ents:
                drawing.append(DXF.Face(**e))



import bpy
import bmesh

from . import mathematics
from . import curves



class LoftedSplineSurface:
    def __init__(self, activeSpline, otherSpline, bMesh, vert0Index, resolution):
        self.splineA = activeSpline
        self.splineO = otherSpline

        self.bMesh = bMesh
        self.vert0Index = vert0Index
        self.resolution = resolution


    def Apply(self, worldMatrixA, worldMatrixO):
        #deltaPar = 1.0 / float(self.resolution - 1)

        par = 0.0
        pointA = worldMatrixA @ self.splineA.CalcPoint(par)
        pointO = worldMatrixO @ self.splineO.CalcPoint(par)
        self.bMesh.verts[self.vert0Index].co = pointA
        self.bMesh.verts[self.vert0Index + 1].co = pointO

        fltResm1 = float(self.resolution - 1)
        for i in range(1, self.resolution):
            par = float(i) / fltResm1

            pointA = worldMatrixA @ self.splineA.CalcPoint(par)
            pointO = worldMatrixO @ self.splineO.CalcPoint(par)
            self.bMesh.verts[self.vert0Index + 2 * i].co = pointA
            self.bMesh.verts[self.vert0Index + 2 * i + 1].co = pointO


    def AddFaces(self):
        currIndexA = self.vert0Index
        currIndexO = self.vert0Index + 1

        bmVerts = self.bMesh.verts
        bmVerts.ensure_lookup_table()

        for i in range(1, self.resolution):
            nextIndexA = self.vert0Index + 2 * i
            nextIndexO = nextIndexA + 1

            self.bMesh.faces.new([bmVerts[currIndexA], bmVerts[currIndexO], bmVerts[nextIndexO], bmVerts[nextIndexA]])

            currIndexA = nextIndexA
            currIndexO = nextIndexO


class LoftedSurface:
    @staticmethod
    def FromSelection():
        selObjects = bpy.context.selected_objects
        if len(selObjects) != 2: raise Exception("len(selObjects) != 2") # shouldn't be possible

        blenderActiveCurve = bpy.context.active_object
        blenderOtherCurve = selObjects[0]
        if blenderActiveCurve == blenderOtherCurve: blenderOtherCurve = selObjects[1]

        aCurve = curves.Curve(blenderActiveCurve)
        oCurve = curves.Curve(blenderOtherCurve)

        name = "TODO: autoname"

        return LoftedSurface(aCurve, oCurve, name)


    def __init__(self, activeCurve, otherCurve, name = "LoftedSurface"):
        self.curveA = activeCurve
        self.curveO = otherCurve
        self.name  = name

        self.nrSplines = self.curveA.nrSplines
        if self.curveO.nrSplines < self.nrSplines: self.nrSplines = self.curveO.nrSplines

        self.bMesh = bmesh.new()

        self.splineSurfaces = self.SetupSplineSurfaces()

        self.Apply()


    def SetupSplineSurfaces(self):
        rvSplineSurfaces = []

        currV0Index = 0
        for i in range(self.nrSplines):
            splineA = self.curveA.splines[i]
            splineO = self.curveO.splines[i]

            res = splineA.resolution
            if splineO.resolution < res: res = splineO.resolution

            for iv in range(2 * res): self.bMesh.verts.new()

            splSurf = LoftedSplineSurface(splineA, splineO, self.bMesh, currV0Index, res)
            splSurf.AddFaces()
            rvSplineSurfaces.append(splSurf)

            currV0Index += 2 * res

        return rvSplineSurfaces


    def Apply(self):
        for splineSurface in self.splineSurfaces: splineSurface.Apply(self.curveA.worldMatrix, self.curveO.worldMatrix)


    def AddToScene(self):
        mesh = bpy.data.meshes.new("Mesh" + self.name)

        self.bMesh.to_mesh(mesh)
        mesh.update()

        meshObject = bpy.data.objects.new(self.name, mesh)

        bpy.context.collection.objects.link(meshObject)



# active spline is swept over other spline (rail)
class SweptSplineSurface:
    def __init__(self, activeSpline, otherSpline, bMesh, vert0Index, resolutionA, resolutionO):
        self.splineA = activeSpline
        self.splineO = otherSpline

        self.bMesh = bMesh
        self.vert0Index = vert0Index
        self.resolutionA = resolutionA
        self.resolutionO = resolutionO


    def Apply(self, worldMatrixA, worldMatrixO):
        localPointsA = []
        fltResAm1 = float(self.resolutionA - 1)
        for i in range(self.resolutionA):
            par = float(i) / fltResAm1
            pointA = self.splineA.CalcPoint(par)
            localPointsA.append(pointA)


        worldPointsO = []
        localDerivativesO = []
        fltResOm1 = float(self.resolutionO - 1)
        for i in range(self.resolutionO):
            par = float(i) / fltResOm1

            pointO = self.splineO.CalcPoint(par)
            worldPointsO.append(worldMatrixO @ pointO)

            derivativeO = self.splineO.CalcDerivative(par)
            localDerivativesO.append(derivativeO)


        currWorldMatrixA = worldMatrixA
        worldMatrixOInv = worldMatrixO.inverted()
        prevDerivativeO = localDerivativesO[0]
        for iO in range(self.resolutionO):
            currDerivativeO = localDerivativesO[iO]
            localRotMatO = mathematics.CalcRotationMatrix(prevDerivativeO, currDerivativeO)

            currLocalAToLocalO = worldMatrixOInv @ currWorldMatrixA
            worldPointsA = []
            for iA in range(self.resolutionA):
                pointALocalToO = currLocalAToLocalO @ localPointsA[iA]
                rotatedPointA = localRotMatO @ pointALocalToO
                worldPointsA.append(worldMatrixO @ rotatedPointA)

            worldOffsetsA = []
            worldPoint0A = worldPointsA[0]
            for i in range(self.resolutionA): worldOffsetsA.append(worldPointsA[i] - worldPoint0A)


            for iA in range(self.resolutionA):
                iVert = self.vert0Index + (self.resolutionA * iO) + iA
                currVert = worldPointsO[iO] + worldOffsetsA[iA]
                self.bMesh.verts[iVert].co = currVert

            prevDerivativeO = currDerivativeO
            currWorldMatrixA = worldMatrixO @ localRotMatO @ currLocalAToLocalO


    def AddFaces(self):
        bmVerts = self.bMesh.verts
        bmVerts.ensure_lookup_table()

        for iO in range(self.resolutionO - 1):
            for iA in range(self.resolutionA - 1):
                currIndexA1 = self.vert0Index + (self.resolutionA * iO) + iA
                currIndexA2 = currIndexA1 + 1
                nextIndexA1 = self.vert0Index + (self.resolutionA * (iO + 1)) + iA
                nextIndexA2 = nextIndexA1 + 1

                self.bMesh.faces.new([bmVerts[currIndexA1], bmVerts[currIndexA2], bmVerts[nextIndexA2], bmVerts[nextIndexA1]])



class SweptSurface:
    @staticmethod
    def FromSelection():
        selObjects = bpy.context.selected_objects
        if len(selObjects) != 2: raise Exception("len(selObjects) != 2") # shouldn't be possible

        blenderActiveCurve = bpy.context.active_object
        blenderOtherCurve = selObjects[0]
        if blenderActiveCurve == blenderOtherCurve: blenderOtherCurve = selObjects[1]

        aCurve = curves.Curve(blenderActiveCurve)
        oCurve = curves.Curve(blenderOtherCurve)

        name = "TODO: autoname"

        return SweptSurface(aCurve, oCurve, name)


    def __init__(self, activeCurve, otherCurve, name = "SweptSurface"):
        self.curveA = activeCurve
        self.curveO = otherCurve
        self.name  = name

        self.nrSplines = self.curveA.nrSplines
        if self.curveO.nrSplines < self.nrSplines: self.nrSplines = self.curveO.nrSplines

        self.bMesh = bmesh.new()

        self.splineSurfaces = self.SetupSplineSurfaces()

        self.Apply()


    def SetupSplineSurfaces(self):
        rvSplineSurfaces = []

        currV0Index = 0
        for i in range(self.nrSplines):
            splineA = self.curveA.splines[i]
            splineO = self.curveO.splines[i]

            resA = splineA.resolution
            resO = splineO.resolution

            for iv in range(resA * resO): self.bMesh.verts.new()

            splSurf = SweptSplineSurface(splineA, splineO, self.bMesh, currV0Index, resA, resO)
            splSurf.AddFaces()
            rvSplineSurfaces.append(splSurf)

            currV0Index += resA * resO

        return rvSplineSurfaces


    def Apply(self):
        for splineSurface in self.splineSurfaces: splineSurface.Apply(self.curveA.worldMatrix, self.curveO.worldMatrix)


    def AddToScene(self):
        mesh = bpy.data.meshes.new("Mesh" + self.name)

        self.bMesh.to_mesh(mesh)
        mesh.update()

        meshObject = bpy.data.objects.new(self.name, mesh)

        bpy.context.collection.objects.link(meshObject)



# profileSpline is swept over rail1Spline and scaled/rotated to have its endpoint on rail2Spline
class BirailedSplineSurface:
    def __init__(self, rail1Spline, rail2Spline, profileSpline, bMesh, vert0Index, resolutionRails, resolutionProfile):
        self.rail1Spline = rail1Spline
        self.rail2Spline = rail2Spline
        self.profileSpline = profileSpline

        self.bMesh = bMesh
        self.vert0Index = vert0Index
        self.resolutionRails = resolutionRails
        self.resolutionProfile = resolutionProfile


    def Apply(self, worldMatrixRail1, worldMatrixRail2, worldMatrixProfile):
        localPointsProfile = []
        fltResProfilem1 = float(self.resolutionProfile - 1)
        for i in range(self.resolutionProfile):
            par = float(i) / fltResProfilem1
            pointProfile = self.profileSpline.CalcPoint(par)
            localPointsProfile.append(pointProfile)


        worldPointsRail1 = []
        localDerivativesRail1 = []
        worldPointsRail2 = []
        fltResRailsm1 = float(self.resolutionRails - 1)
        for i in range(self.resolutionRails):
            par = float(i) / fltResRailsm1

            pointRail1 = self.rail1Spline.CalcPoint(par)
            worldPointsRail1.append(worldMatrixRail1 @ pointRail1)

            derivativeRail1 = self.rail1Spline.CalcDerivative(par)
            localDerivativesRail1.append(derivativeRail1)

            pointRail2 = self.rail2Spline.CalcPoint(par)
            worldPointsRail2.append(worldMatrixRail2 @ pointRail2)


        currWorldMatrixProfile = worldMatrixProfile
        worldMatrixRail1Inv = worldMatrixRail1.inverted()
        prevDerivativeRail1 = localDerivativesRail1[0]
        for iRail in range(self.resolutionRails):
            currDerivativeRail1 = localDerivativesRail1[iRail]
            localRotMatRail1 = mathematics.CalcRotationMatrix(prevDerivativeRail1, currDerivativeRail1)

            currLocalProfileToLocalRail1 = worldMatrixRail1Inv @ currWorldMatrixProfile
            worldPointsProfileRail1 = []
            for iProfile in range(self.resolutionProfile):
                pointProfileLocalToRail1 = currLocalProfileToLocalRail1 @ localPointsProfile[iProfile]
                rotatedPointProfile = localRotMatRail1 @ pointProfileLocalToRail1
                worldPointsProfileRail1.append(worldMatrixRail1 @ rotatedPointProfile)

            worldOffsetsProfileRail1 = []
            worldPoint0ProfileRail1 = worldPointsProfileRail1[0]
            for iProfile in range(self.resolutionProfile): worldOffsetsProfileRail1.append(worldPointsProfileRail1[iProfile] - worldPoint0ProfileRail1)

            worldStartPointProfileRail1 = worldPointsRail1[iRail]
            worldEndPointProfileRail1 = worldStartPointProfileRail1 + worldOffsetsProfileRail1[-1]
            v3From = worldEndPointProfileRail1 - worldStartPointProfileRail1
            v3To = worldPointsRail2[iRail] - worldStartPointProfileRail1
            if not v3From.magnitude == 0:
                scaleFactorRail2 = v3To.magnitude / v3From.magnitude
            else:
                scaleFactorRail2 = 1
            rotMatRail2 = mathematics.CalcRotationMatrix(v3From, v3To)

            worldOffsetsProfileRail2 = []
            for iProfile in range(self.resolutionProfile):
                offsetProfileRail1 = worldOffsetsProfileRail1[iProfile]
                worldOffsetsProfileRail2.append(rotMatRail2 @ (offsetProfileRail1 * scaleFactorRail2))


            for iProfile in range(self.resolutionProfile):
                iVert = self.vert0Index + (self.resolutionProfile * iRail) + iProfile
                currVert = worldPointsRail1[iRail] + worldOffsetsProfileRail2[iProfile]
                self.bMesh.verts[iVert].co = currVert

            prevDerivativeRail1 = currDerivativeRail1
            currWorldMatrixProfile = worldMatrixRail1 @ localRotMatRail1 @ currLocalProfileToLocalRail1


    def AddFaces(self):
        bmVerts = self.bMesh.verts
        bmVerts.ensure_lookup_table()

        for iRail in range(self.resolutionRails - 1):
            for iProfile in range(self.resolutionProfile - 1):
                currIndex1 = self.vert0Index + (self.resolutionProfile * iRail) + iProfile
                currIndex2 = currIndex1 + 1
                nextIndex1 = self.vert0Index + (self.resolutionProfile * (iRail + 1)) + iProfile
                nextIndex2 = nextIndex1 + 1

                self.bMesh.faces.new([bmVerts[currIndex1], bmVerts[currIndex2], bmVerts[nextIndex2], bmVerts[nextIndex1]])



class BirailedSurface:
    @staticmethod
    def FromSelection():
        selectedObjects = bpy.context.selected_objects

        rail1Curve = curves.Curve(selectedObjects[0])
        rail2Curve = curves.Curve(selectedObjects[1])
        profileCurve = curves.Curve(selectedObjects[2])

        name = "BirailedSurface"

        return BirailedSurface(rail1Curve, rail2Curve, profileCurve, name)


    def __init__(self, rail1Curve, rail2Curve, profileCurve, name = "BirailedSurface"):
        self.rail1Curve = rail1Curve
        self.rail2Curve = rail2Curve
        self.profileCurve = profileCurve
        self.name  = name

        self.nrSplines = self.rail1Curve.nrSplines
        if self.rail2Curve.nrSplines < self.nrSplines: self.nrSplines = self.rail2Curve.nrSplines
        if self.profileCurve.nrSplines < self.nrSplines: self.nrSplines = self.profileCurve.nrSplines

        self.bMesh = bmesh.new()

        self.splineSurfaces = self.SetupSplineSurfaces()

        self.Apply()


    def SetupSplineSurfaces(self):
        rvSplineSurfaces = []

        currV0Index = 0
        for i in range(self.nrSplines):
            splineRail1 = self.rail1Curve.splines[i]
            splineRail2 = self.rail2Curve.splines[i]
            splineProfile = self.profileCurve.splines[i]

            resProfile = splineProfile.resolution
            resRails = splineRail1.resolution
            if splineRail2.resolution < resRails: resRails = splineRail2.resolution

            for iv in range(resProfile * resRails): self.bMesh.verts.new()

            splSurf = BirailedSplineSurface(splineRail1, splineRail2, splineProfile, self.bMesh, currV0Index, resRails, resProfile)
            splSurf.AddFaces()
            rvSplineSurfaces.append(splSurf)

            currV0Index += resProfile * resRails

        return rvSplineSurfaces


    def Apply(self):
        for splineSurface in self.splineSurfaces: splineSurface.Apply(self.rail1Curve.worldMatrix, self.rail2Curve.worldMatrix, self.profileCurve.worldMatrix)


    def AddToScene(self):
        mesh = bpy.data.meshes.new("Mesh" + self.name)

        self.bMesh.to_mesh(mesh)
        mesh.update()

        meshObject = bpy.data.objects.new(self.name, mesh)

        bpy.context.collection.objects.link(meshObject)



#!/usr/bin/env python3

import argparse
import pathlib
import requests
import shutil
import subprocess
from typing import Callable, Iterator, List, Tuple

# supported archive and platform endings, used to create actual archive names
archive_endings = ["windows64.zip", "linux64.tar.xz", "macOS.dmg"]


def add_optional_argument(option: str, help: str) -> None:
    global parser
    """Add an optional argument

    Args:
        option (str): Option to add
        help (str): Help description for the argument
    """
    parser.add_argument(option, help=help, action='store_const', const=1)


def blender_archives(version: str) -> Iterator[str]:
    """Generator for Blender archives for version.

    Yields for items in archive_endings an archive name in the form of
    blender-{version}-{ending}.

    Args:
        version (str): Version string of the form 2.83.2


    Yields:
        Iterator[str]: Name in the form of blender-{version}-{ending}
    """
    global archive_endings

    for ending in archive_endings:
        yield f"blender-{version}-{ending}"


def get_archive_type(archive_type: str, version: str) -> str:
    """Return the archive of given type and version.

    Args:
        archive_type (str): extension for archive type to check for
        version (str): Version string in the form 2.83.2

    Raises:
        Exception: Execption when archive type isn't found

    Returns:
        str: archive name for given type
    """

    for archive in blender_archives(version):
        if archive.endswith(archive_type):
            return archive
    raise Exception("Unknown archive type")


def execute_command(cmd: List[str], name: str, errcode: int, cwd=".", capture_output=True) -> str:
    """Execute the given command.

    Returns the process stdout upon success if any.

    On error print message the command with name that has failed. Print stdout
    and stderr of the process if any, and then exit with given error code.

    Args:
        cmd (List[str]): Command in list format, each argument as their own item
        name (str): Name of command to use when printing to command-line
        errcode (int): Error code to use in case of exit()
        cwd (str, optional): Folder to use as current work directory for command
                             execution. Defaults to ".".
        capture_output (bool, optional): Whether to capture command output or not.
                                         Defaults to True.

    Returns:
        str: stdout if any, or empty string
    """
    cmd_process = subprocess.run(
        cmd, capture_output=capture_output, encoding="UTF-8", cwd=cwd)
    if cmd_process.returncode == 0:
        if cmd_process.stdout:
            return cmd_process.stdout
        else:
            return ""
    else:
        print(f"ERROR: {name} failed.")
        if cmd_process.stdout:
            print(cmd_process.stdout)
        if cmd_process.stderr:
            print(cmd_process.stderr)
        exit(errcode)
        return ""


def download_archives(base_url: str, archives: Callable[[str], Iterator[str]], version: str, dst_dir: pathlib.Path):
    """Download archives from the given base_url.

    Archives is a generator for Blender archive names based on version.

    Archive names are appended to the base_url to load from, and appended to
    dst_dir to save to.

    Args:
        base_url (str): Base URL to load archives from
        archives (Callable[[str], Iterator[str]]): Generator for Blender archive
                                                   names based on version
        version (str): Version string in the form of 2.83.2
        dst_dir (pathlib.Path): Download destination
    """

    if base_url[-1] != '/':
        base_url = base_url + '/'

    for archive in archives(version):
        download_url = f"{base_url}{archive}"
        target_file = dst_dir.joinpath(archive)
        download_file(download_url, target_file)


def download_file(from_url: str, to_file: pathlib.Path) -> None:
    """Download from_url as to_file.

    Actual downloading will be skipped if --skipdl is given on the command-line.

    Args:
        from_url (str): Full URL to resource to download
        to_file (pathlib.Path): Full path to save downloaded resource as
    """
    global args

    if not args.skipdl or not to_file.exists():
        print(f"Downloading {from_url}")
        with open(to_file, "wb") as download_zip:
            response = requests.get(from_url)
            if response.status_code != requests.codes.ok:
                print(f"ERROR: failed to download {from_url} (status code: {response.status_code})")
                exit(1313)
            download_zip.write(response.content)
    else:
        print(f"Downloading {from_url} skipped")
    print("   ... OK")


def copy_contents_from_dmg_to_path(dmg_file: pathlib.Path, dst: pathlib.Path) -> None:
    """Copy the contents of the given DMG file to the destination folder.

    Args:
        dmg_file (pathlib.Path): Full path to DMG archive to extract from
        dst (pathlib.Path): Full path to destination to extract to
    """
    hdiutil_attach = ["hdiutil",
                      "attach",
                      "-readonly",
                      f"{dmg_file}"
                      ]
    attached = execute_command(hdiutil_attach, "hdiutil attach", 1)

    # Last line of output is what we want, it is of the form
    # /dev/somedisk    Apple_HFS     /Volumes/Blender
    # We want to retain the mount point, and the folder the mount is
    # created on. The mounted disk we need for detaching, the folder we
    # need to be able to copy the contents to where we can use them
    attachment_items = attached.splitlines()[-1].split()
    mounted_disk = attachment_items[0]
    source_location = pathlib.Path(attachment_items[2], "Blender.app")

    print(f"{source_location} -> {dst}")

    shutil.copytree(source_location, dst)

    hdiutil_detach = ["hdiutil",
                      "detach",
                      f"{mounted_disk}"
                      ]
    execute_command(hdiutil_detach, "hdiutil detach", 2)


def create_build_script(template_name: str, vars: List[Tuple[str, str]]) -> pathlib.Path:
    """
    Create the Steam build script

    Use the given template and template variable tuple list.

    Returns pathlib.Path to the created script.

    Args:
        template_name (str): [description]
        vars (List[Tuple[str, str]]): [description]

    Returns:
        pathlib.Path: Full path to the generated script
    """
    build_script = pathlib.Path(".", template_name).read_text()
    for var in vars:
        build_script = build_script.replace(var[0], var[1])
    build_script_file = template_name.replace(".template", "")
    build_script_path = pathlib.Path(".", build_script_file)
    build_script_path.write_text(build_script)
    return build_script_path


def clean_up() -> None:
    """Remove intermediate files depending on given command-line arguments
    """
    global content_location, args

    if not args.leavearch and not args.leaveextracted:
        shutil.rmtree(content_location)

    if args.leavearch and not args.leaveextracted:
        shutil.rmtree(content_location.joinpath(zip_extract_folder))
        shutil.rmtree(content_location.joinpath(tarxz_extract_folder))
        shutil.rmtree(content_location.joinpath(dmg_extract_folder))

    if args.leaveextracted and not args.leavearch:
        import os
        os.remove(content_location.joinpath(zipped_blender))
        os.remove(content_location.joinpath(tarxz_blender))
        os.remove(content_location.joinpath(dmg_blender))


def extract_archive(archive: str, extract_folder_name: str,
                    cmd: List[str], errcode: int) -> None:
    """Extract all files from archive to given folder name.

    Will not extract if
    target folder already exists, or if --skipextract was given on the
    command-line.

    Args:
        archive (str): Archive name to extract
        extract_folder_name (str): Folder name to extract to
        cmd (List[str]): Command with arguments to use
        errcode (int): Error code to use for exit()
    """
    global args, content_location

    extract_location = content_location.joinpath(extract_folder_name)

    pre_extract = set(content_location.glob("*"))

    if not args.skipextract or not extract_location.exists():
        print(f"Extracting files from {archive}...")
        cmd.append(content_location.joinpath(archive))
        execute_command(cmd, cmd[0], errcode, cwd=content_location)
        # in case we use a non-release archive the naming will be incorrect.
        # simply rename to expected target name
        post_extract = set(content_location.glob("*"))
        diff_extract = post_extract - pre_extract
        if not extract_location in diff_extract:
            folder_to_rename = list(diff_extract)[0]
            folder_to_rename.rename(extract_location)
        print("   OK")
    else:
        print(f"Skipping extraction {archive}!")

# ==============================================================================


parser = argparse.ArgumentParser()

parser.add_argument("--baseurl", required=True,
                    help="The base URL for files to download, "
                    "i.e. https://download.blender.org/release/Blender2.83/")

parser.add_argument("--version", required=True,
                    help="The Blender version to release, in the form 2.83.3")

parser.add_argument("--appid", required=True,
                    help="The Blender App ID on Steam")
parser.add_argument("--winid", required=True,
                    help="The Windows depot ID")
parser.add_argument("--linuxid", required=True,
                    help="The Linux depot ID")
parser.add_argument("--macosid", required=True,
                    help="The MacOS depot ID")

parser.add_argument("--steamcmd", required=True,
                    help="Path to the steamcmd")
parser.add_argument("--steamuser", required=True,
                    help="The login for the Steam builder user")
parser.add_argument("--steampw", required=True,
                    help="Login password for the Steam builder user")

add_optional_argument("--dryrun",
                      "If set the Steam files will not be uploaded")
add_optional_argument("--leavearch",
                      help="If set don't clean up the downloaded archives")
add_optional_argument("--leaveextracted",
                      help="If set don't clean up the extraction folders")
add_optional_argument("--skipdl",
                      help="If set downloading the archives is skipped if it already exists locally.")
add_optional_argument("--skipextract",
                      help="If set skips extracting of archives. The tool assumes the archives"
                      "have already been extracted to their correct locations")

args = parser.parse_args()

VERSIONNODOTS = args.version.replace('.', '')
OUTPUT = f"output{VERSIONNODOTS}"
CONTENT = f"content{VERSIONNODOTS}"

# ===== set up main locations

content_location = pathlib.Path(".", CONTENT).absolute()
output_location = pathlib.Path(".", OUTPUT).absolute()

content_location.mkdir(parents=True, exist_ok=True)
output_location.mkdir(parents=True, exist_ok=True)

# ===== login

# Logging into Steam once to ensure the SDK updates itself properly. If we don't
# do that the combined +login and +run_app_build_http at the end of the tool
# will fail.
steam_login = [args.steamcmd,
               "+login",
               args.steamuser,
               args.steampw,
               "+quit"
               ]
print("Logging in to Steam...")
execute_command(steam_login, "Login to Steam", 10)
print("   OK")

# ===== prepare Steam build scripts

template_vars = [
    ("[APPID]", args.appid),
    ("[OUTPUT]", OUTPUT),
    ("[CONTENT]", CONTENT),
    ("[VERSION]", args.version),
    ("[WINID]", args.winid),
    ("[LINUXID]", args.linuxid),
    ("[MACOSID]", args.macosid),
    ("[DRYRUN]", f"{args.dryrun}" if args.dryrun else "0")
]

blender_app_build = create_build_script(
    "blender_app_build.vdf.template", template_vars)
create_build_script("depot_build_win.vdf.template", template_vars)
create_build_script("depot_build_linux.vdf.template", template_vars)
create_build_script("depot_build_macos.vdf.template", template_vars)

# ===== download archives

download_archives(args.baseurl, blender_archives,
                  args.version, content_location)

# ===== set up file and folder names

zipped_blender = get_archive_type("zip", args.version)
zip_extract_folder = zipped_blender.replace(".zip", "")
tarxz_blender = get_archive_type("tar.xz", args.version)
tarxz_extract_folder = tarxz_blender.replace(".tar.xz", "")
dmg_blender = get_archive_type("dmg", args.version)
dmg_extract_folder = dmg_blender.replace(".dmg", "")

# ===== extract

unzip_cmd = ["unzip", "-q"]
extract_archive(zipped_blender, zip_extract_folder,  unzip_cmd, 3)

untarxz_cmd = ["tar", "-xf"]
extract_archive(tarxz_blender, tarxz_extract_folder, untarxz_cmd, 4)

if not args.skipextract or not content_location.joinpath(dmg_extract_folder).exists():
    print("Extracting files from Blender MacOS archive...")
    blender_dmg = content_location.joinpath(dmg_blender)
    target_location = content_location.joinpath(
        dmg_extract_folder, "Blender.app")
    copy_contents_from_dmg_to_path(blender_dmg, target_location)
    print("   OK")
else:
    print("Skipping extraction of .dmg!")

# ===== building

print("Build Steam game files...")
steam_build = [args.steamcmd,
               "+login",
               args.steamuser,
               args.steampw,
               "+run_app_build_http",
               blender_app_build.absolute(),
               "+quit"
               ]
execute_command(steam_build, "Build with steamcmd", 13)
print("   OK")

clean_up()



#!/usr/bin/env python3

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

import logging.config
from pathlib import Path
from typing import List

from codesign.macos_code_signer import MacOSCodeSigner
import codesign.config_server

if __name__ == "__main__":
    entitlements_file = codesign.config_server.MACOS_ENTITLEMENTS_FILE
    if not entitlements_file.exists():
        raise SystemExit(
            'Entitlements file {entitlements_file} does not exist.')
    if not entitlements_file.is_file():
        raise SystemExit(
            'Entitlements file {entitlements_file} is not a file.')

    logging.config.dictConfig(codesign.config_server.LOGGING)
    code_signer = MacOSCodeSigner(codesign.config_server)
    code_signer.run_signing_server()



# -*- coding: utf-8 -*-
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


import bpy

import time
import copy

from mathutils import (
        Euler,
        Matrix,
        Vector,
        )
from math import pi, sin, degrees, radians, atan2, copysign, cos, acos
from math import floor
from random import random, uniform, seed, choice, getstate, setstate, randint
from collections import deque, OrderedDict

tau = 2 * pi

# Initialise the split error and axis vectors
splitError = 0.0
zAxis = Vector((0, 0, 1))
yAxis = Vector((0, 1, 0))
xAxis = Vector((1, 0, 0))


# This class will contain a part of the tree which needs to be extended and the required tree parameters
class stemSpline:
    def __init__(self, spline, curvature, curvatureV, attractUp, segments, maxSegs,
                 segLength, childStems, stemRadStart, stemRadEnd, splineNum, ofst, pquat):
        self.spline = spline
        self.p = spline.bezier_points[-1]
        self.curv = curvature
        self.curvV = curvatureV
        self.vertAtt = attractUp
        self.seg = segments
        self.segMax = maxSegs
        self.segL = segLength
        self.children = childStems
        self.radS = stemRadStart
        self.radE = stemRadEnd
        self.splN = splineNum
        self.offsetLen = ofst
        self.patentQuat = pquat
        self.curvSignx = 1
        self.curvSigny = 1

    # This method determines the quaternion of the end of the spline
    def quat(self):
        if len(self.spline.bezier_points) == 1:
            return ((self.spline.bezier_points[-1].handle_right -
                     self.spline.bezier_points[-1].co).normalized()).to_track_quat('Z', 'Y')
        else:
            return ((self.spline.bezier_points[-1].co -
                     self.spline.bezier_points[-2].co).normalized()).to_track_quat('Z', 'Y')

    # Determine the declination
    def dec(self):
        tempVec = zAxis.copy()
        tempVec.rotate(self.quat())
        return zAxis.angle(tempVec)

    # Update the end of the spline and increment the segment count
    def updateEnd(self):
        self.p = self.spline.bezier_points[-1]
        self.seg += 1


# This class contains the data for a point where a new branch will sprout
class childPoint:
    def __init__(self, coords, quat, radiusPar, offset, sOfst, lengthPar, parBone):
        self.co = coords
        self.quat = quat
        self.radiusPar = radiusPar
        self.offset = offset
        self.stemOffset = sOfst
        self.lengthPar = lengthPar
        self.parBone = parBone


# This function calculates the shape ratio as defined in the paper
def shapeRatio(shape, ratio, pruneWidthPeak=0.0, prunePowerHigh=0.0, prunePowerLow=0.0, custom=None):
    if shape == 0:
        return 0.05 + 0.95 * ratio  # 0.2 + 0.8 * ratio
    elif shape == 1:
        return 0.2 + 0.8 * sin(pi * ratio)
    elif shape == 2:
        return 0.2 + 0.8 * sin(0.5 * pi * ratio)
    elif shape == 3:
        return 1.0
    elif shape == 4:
        return 0.5 + 0.5 * ratio
    elif shape == 5:
        if ratio <= 0.7:
            return 0.05 + 0.95 * ratio / 0.7
        else:
            return 0.05 + 0.95 * (1.0 - ratio) / 0.3
    elif shape == 6:
        return 1.0 - 0.8 * ratio
    elif shape == 7:
        if ratio <= 0.7:
            return 0.5 + 0.5 * ratio / 0.7
        else:
            return 0.5 + 0.5 * (1.0 - ratio) / 0.3
    elif shape == 8:
        r = 1 - ratio
        if r == 1:
            v = custom[3]
        elif r >= custom[2]:
            pos = (r - custom[2]) / (1 - custom[2])
            # if (custom[0] >= custom[1] <= custom[3]) or (custom[0] <= custom[1] >= custom[3]):
            pos = pos * pos
            v = (pos * (custom[3] - custom[1])) + custom[1]
        else:
            pos = r / custom[2]
            # if (custom[0] >= custom[1] <= custom[3]) or (custom[0] <= custom[1] >= custom[3]):
            pos = 1 - (1 - pos) * (1 - pos)
            v = (pos * (custom[1] - custom[0])) + custom[0]

        return v

    elif shape == 9:
        if (ratio < (1 - pruneWidthPeak)) and (ratio > 0.0):
            return ((ratio / (1 - pruneWidthPeak))**prunePowerHigh)
        elif (ratio >= (1 - pruneWidthPeak)) and (ratio < 1.0):
            return (((1 - ratio) / pruneWidthPeak)**prunePowerLow)
        else:
            return 0.0

    elif shape == 10:
        return 0.5 + 0.5 * (1 - ratio)


# This function determines the actual number of splits at a given point using the global error
def splits(n):
    global splitError
    nEff = round(n + splitError, 0)
    splitError -= (nEff - n)
    return int(nEff)


def splits2(n):
    r = random()
    if r < n:
        return 1
    else:
        return 0


def splits3(n):
    ni = int(n)
    nf = n - int(n)
    r = random()
    if r < nf:
        return ni + 1
    else:
        return ni + 0


# Determine the declination from a given quaternion
def declination(quat):
    tempVec = zAxis.copy()
    tempVec.rotate(quat)
    tempVec.normalize()
    return degrees(acos(tempVec.z))


# Determines the angle of upward rotation of a segment due to attractUp
def curveUp(attractUp, quat, curveRes):
    tempVec = yAxis.copy()
    tempVec.rotate(quat)
    tempVec.normalize()

    dec = radians(declination(quat))
    curveUpAng = attractUp * dec * abs(tempVec.z) / curveRes
    if (-dec + curveUpAng) < -pi:
        curveUpAng = -pi + dec
    if (dec - curveUpAng) < 0:
        curveUpAng = dec
    return curveUpAng


# Evaluate a bezier curve for the parameter 0<=t<=1 along its length
def evalBez(p1, h1, h2, p2, t):
    return ((1 - t)**3) * p1 + (3 * t * (1 - t)**2) * h1 + (3 * (t**2) * (1 - t)) * h2 + (t**3) * p2


# Evaluate the unit tangent on a bezier curve for t
def evalBezTan(p1, h1, h2, p2, t):
    return (
            (-3 * (1 - t)**2) * p1 + (-6 * t * (1 - t) + 3 * (1 - t)**2) * h1 +
            (-3 * (t**2) + 6 * t * (1 - t)) * h2 + (3 * t**2) * p2
            ).normalized()


# Determine the range of t values along a splines length where child stems are formed
def findChildPoints(stemList, numChild):
    numPoints = sum([len(n.spline.bezier_points) for n in stemList])
    numSplines = len(stemList)
    numSegs = numPoints - numSplines
    numPerSeg = numChild / numSegs
    numMain = round(numPerSeg * stemList[0].segMax, 0)
    return [(a + 1) / (numMain) for a in range(int(numMain))]


def findChildPoints2(stemList, numChild):
    return [(a + 1) / (numChild) for a in range(int(numChild))]


# Find the coordinates, quaternion and radius for each t on the stem
def interpStem1(stem, tVals, lPar, parRad):
    points = stem.spline.bezier_points
    numPoints = len(points)
    checkVal = (stem.segMax - (numPoints - 1)) / stem.segMax
    # Loop through all the parametric values to be determined
    tempList = deque()
    for t in tVals:
        if t == 1.0:
            index = numPoints - 2
            coord = points[-1].co
            quat = (points[-1].handle_right - points[-1].co).to_track_quat('Z', 'Y')
            radius = points[-1].radius

            tempList.append(
                    childPoint(coord, quat, (parRad, radius), t, lPar, 'bone' +
                              (str(stem.splN).rjust(3, '0')) + '.' + (str(index).rjust(3, '0')))
                    )

        elif (t >= checkVal) and (t < 1.0):
            scaledT = (t - checkVal) / ((1 - checkVal) + .0001)
            length = (numPoints - 1) * scaledT
            index = int(length)

            tTemp = length - index
            coord = evalBez(
                        points[index].co, points[index].handle_right,
                        points[index + 1].handle_left, points[index + 1].co, tTemp
                        )
            quat = (
                evalBezTan(
                    points[index].co, points[index].handle_right,
                    points[index + 1].handle_left, points[index + 1].co, tTemp)
                    ).to_track_quat('Z', 'Y')
            # Not sure if this is the parent radius at the child point or parent start radius
            radius = (1 - tTemp) * points[index].radius + tTemp * points[index + 1].radius

            tempList.append(
                    childPoint(
                            coord, quat, (parRad, radius), t, lPar, 'bone' +
                            (str(stem.splN).rjust(3, '0')) + '.' + (str(index).rjust(3, '0')))
                            )
    return tempList


def interpStem(stem, tVals, lPar, parRad, maxOffset, baseSize):
    points = stem.spline.bezier_points
    numSegs = len(points) - 1
    stemLen = stem.segL * numSegs

    checkBottom = stem.offsetLen / maxOffset
    checkTop = checkBottom + (stemLen / maxOffset)

    # Loop through all the parametric values to be determined
    tempList = deque()
    for t in tVals:
        if (t >= checkBottom) and (t <= checkTop) and (t < 1.0):
            scaledT = (t - checkBottom) / (checkTop - checkBottom)
            ofst = ((t - baseSize) / (checkTop - baseSize)) * (1 - baseSize) + baseSize

            length = numSegs * scaledT
            index = int(length)
            tTemp = length - index

            coord = evalBez(
                        points[index].co, points[index].handle_right,
                        points[index + 1].handle_left, points[index + 1].co, tTemp
                        )
            quat = (
                evalBezTan(
                    points[index].co, points[index].handle_right,
                    points[index + 1].handle_left, points[index + 1].co, tTemp
                    )
                ).to_track_quat('Z', 'Y')
            # Not sure if this is the parent radius at the child point or parent start radius
            radius = (1 - tTemp) * points[index].radius + tTemp * points[index + 1].radius

            tempList.append(
                    childPoint(
                        coord, quat, (parRad, radius), t, ofst, lPar,
                        'bone' + (str(stem.splN).rjust(3, '0')) + '.' + (str(index).rjust(3, '0')))
                    )

    # add stem at tip
    index = numSegs - 1
    coord = points[-1].co
    quat = (points[-1].handle_right - points[-1].co).to_track_quat('Z', 'Y')
    radius = points[-1].radius
    tempList.append(
                childPoint(
                        coord, quat, (parRad, radius), 1, 1, lPar,
                        'bone' + (str(stem.splN).rjust(3, '0')) + '.' + (str(index).rjust(3, '0'))
                        )
                    )

    return tempList


# round down bone number
def roundBone(bone, step):
    bone_i = bone[:-3]
    bone_n = int(bone[-3:])
    bone_n = int(bone_n / step) * step
    return bone_i + str(bone_n).rjust(3, '0')


# Convert a list of degrees to radians
def toRad(list):
    return [radians(a) for a in list]


def anglemean(a1, a2, fac):
    x1 = sin(a1)
    y1 = cos(a1)
    x2 = sin(a2)
    y2 = cos(a2)
    x = x1 + (x2 - x1) * fac
    y = y1 + (y2 - y1) * fac
    return atan2(x, y)


# This is the function which extends (or grows) a given stem.
def growSpline(n, stem, numSplit, splitAng, splitAngV, splineList,
               hType, splineToBone, closeTip, kp, splitHeight, outAtt, stemsegL,
               lenVar, taperCrown, boneStep, rotate, rotateV):

    # curv at base
    sCurv = stem.curv
    if (n == 0) and (kp <= splitHeight):
        sCurv = 0.0

    # curveangle = sCurv + (uniform(-stem.curvV, stem.curvV) * kp)
    # curveVar = uniform(-stem.curvV, stem.curvV) * kp
    curveangle = sCurv + (uniform(0, stem.curvV) * kp * stem.curvSignx)
    curveVar = uniform(0, stem.curvV) * kp * stem.curvSigny
    stem.curvSignx *= -1
    stem.curvSigny *= -1

    curveVarMat = Matrix.Rotation(curveVar, 3, 'Y')

    # First find the current direction of the stem
    dir = stem.quat()

    if n == 0:
        adir = zAxis.copy()
        adir.rotate(dir)

        ry = atan2(adir[0], adir[2])
        adir.rotate(Euler((0, -ry, 0)))
        rx = atan2(adir[1], adir[2])

        dir = Euler((-rx, ry, 0), 'XYZ')

    # length taperCrown
    if n == 0:
        dec = declination(dir) / 180
        dec = dec ** 2
        tf = 1 - (dec * taperCrown * 30)
        tf = max(.1, tf)
    else:
        tf = 1.0

    # outward attraction
    if (n > 0) and (kp > 0) and (outAtt > 0):
        p = stem.p.co.copy()
        d = atan2(p[0], -p[1]) + tau
        edir = dir.to_euler('XYZ', Euler((0, 0, d), 'XYZ'))
        d = anglemean(edir[2], d, (kp * outAtt))
        dirv = Euler((edir[0], edir[1], d), 'XYZ')
        dir = dirv.to_quaternion()
    """
    # parent weight
    parWeight = kp * degrees(stem.curvV) * pi
    parWeight = parWeight * kp
    parWeight = kp
    if (n > 1) and (parWeight != 0):
        d1 = zAxis.copy()
        d2 = zAxis.copy()
        d1.rotate(dir)
        d2.rotate(stem.patentQuat)

        x = d1[0] + ((d2[0] - d1[0]) * parWeight)
        y = d1[1] + ((d2[1] - d1[1]) * parWeight)
        z = d1[2] + ((d2[2] - d1[2]) * parWeight)

        d3 = Vector((x, y, z))
        dir = d3.to_track_quat('Z', 'Y')
    """

    # If the stem splits, we need to add new splines etc
    if numSplit > 0:
        # Get the curve data
        cuData = stem.spline.id_data.name
        cu = bpy.data.curves[cuData]

        # calc split angles
        angle = choice([-1, 1]) * (splitAng + uniform(-splitAngV, splitAngV))
        if n > 0:
            # make branches flatter
            angle *= max(1 - declination(dir) / 90, 0) * .67 + .33
        spreadangle = choice([-1, 1]) * (splitAng + uniform(-splitAngV, splitAngV))

        # branchRotMat = Matrix.Rotation(radians(uniform(0, 360)), 3, 'Z')
        if not hasattr(stem, 'rLast'):
            stem.rLast = radians(uniform(0, 360))

        br = rotate[0] + uniform(-rotateV[0], rotateV[0])
        branchRot = stem.rLast + br
        branchRotMat = Matrix.Rotation(branchRot, 3, 'Z')
        stem.rLast = branchRot

        # Now for each split add the new spline and adjust the growth direction
        for i in range(numSplit):
            # find split scale
            lenV = uniform(1 - lenVar, 1 + lenVar)
            bScale = min(lenV * tf, 1)

            newSpline = cu.splines.new('BEZIER')
            newPoint = newSpline.bezier_points[-1]
            (newPoint.co, newPoint.handle_left_type, newPoint.handle_right_type) = (stem.p.co, 'VECTOR', 'VECTOR')
            newPoint.radius = (
                        stem.radS * (1 - stem.seg / stem.segMax) + stem.radE * (stem.seg / stem.segMax)
                        ) * bScale
            # Here we make the new "sprouting" stems diverge from the current direction
            divRotMat = Matrix.Rotation(angle + curveangle, 3, 'X')
            dirVec = zAxis.copy()
            dirVec.rotate(divRotMat)

            # horizontal curvature variation
            dirVec.rotate(curveVarMat)

            if n == 0:  # Special case for trunk splits
                dirVec.rotate(branchRotMat)

                ang = pi - ((tau) / (numSplit + 1)) * (i + 1)
                dirVec.rotate(Matrix.Rotation(ang, 3, 'Z'))

            # Spread the stem out in a random fashion
            spreadMat = Matrix.Rotation(spreadangle, 3, 'Y')
            if n != 0:  # Special case for trunk splits
                dirVec.rotate(spreadMat)

            dirVec.rotate(dir)

            # Introduce upward curvature
            upRotAxis = xAxis.copy()
            upRotAxis.rotate(dirVec.to_track_quat('Z', 'Y'))
            curveUpAng = curveUp(stem.vertAtt, dirVec.to_track_quat('Z', 'Y'), stem.segMax)
            upRotMat = Matrix.Rotation(-curveUpAng, 3, upRotAxis)
            dirVec.rotate(upRotMat)

            # Make the growth vec the length of a stem segment
            dirVec.normalize()

            # split length variation
            stemL = stemsegL * lenV
            dirVec *= stemL * tf
            ofst = stem.offsetLen + (stem.segL * (len(stem.spline.bezier_points) - 1))

            # dirVec *= stem.segL

            # Get the end point position
            end_co = stem.p.co.copy()

            # Add the new point and adjust its coords, handles and radius
            newSpline.bezier_points.add(1)
            newPoint = newSpline.bezier_points[-1]
            (newPoint.co, newPoint.handle_left_type, newPoint.handle_right_type) = (end_co + dirVec, hType, hType)
            newPoint.radius = (
                        stem.radS * (1 - (stem.seg + 1) / stem.segMax) +
                        stem.radE * ((stem.seg + 1) / stem.segMax)
                        ) * bScale
            if (stem.seg == stem.segMax - 1) and closeTip:
                newPoint.radius = 0.0
            # If this isn't the last point on a stem, then we need to add it
            # to the list of stems to continue growing
            # print(stem.seg != stem.segMax, stem.seg, stem.segMax)
            # if stem.seg != stem.segMax: # if probs not necessary
            nstem = stemSpline(
                        newSpline, stem.curv, stem.curvV, stem.vertAtt, stem.seg + 1,
                        stem.segMax, stemL, stem.children,
                        stem.radS * bScale, stem.radE * bScale, len(cu.splines) - 1, ofst, stem.quat()
                        )
            nstem.splitlast = 1  # numSplit  # keep track of numSplit for next stem
            nstem.rLast = branchRot + pi
            splineList.append(nstem)
            bone = 'bone' + (str(stem.splN)).rjust(3, '0') + '.' + \
                    (str(len(stem.spline.bezier_points) - 2)).rjust(3, '0')
            bone = roundBone(bone, boneStep[n])
            splineToBone.append((bone, False, True, len(stem.spline.bezier_points) - 2))

        # The original spline also needs to keep growing so adjust its direction too
        divRotMat = Matrix.Rotation(-angle + curveangle, 3, 'X')
        dirVec = zAxis.copy()
        dirVec.rotate(divRotMat)

        # horizontal curvature variation
        dirVec.rotate(curveVarMat)

        if n == 0:  # Special case for trunk splits
            dirVec.rotate(branchRotMat)

        # spread
        spreadMat = Matrix.Rotation(-spreadangle, 3, 'Y')
        if n != 0:  # Special case for trunk splits
            dirVec.rotate(spreadMat)

        dirVec.rotate(dir)

        stem.splitlast = 1  # numSplit #keep track of numSplit for next stem

    else:
        # If there are no splits then generate the growth direction without accounting for spreading of stems
        dirVec = zAxis.copy()
        divRotMat = Matrix.Rotation(curveangle, 3, 'X')
        dirVec.rotate(divRotMat)

        # horizontal curvature variation
        dirVec.rotate(curveVarMat)

        dirVec.rotate(dir)

        stem.splitlast = 0  # numSplit #keep track of numSplit for next stem

    # Introduce upward curvature
    upRotAxis = xAxis.copy()
    upRotAxis.rotate(dirVec.to_track_quat('Z', 'Y'))
    curveUpAng = curveUp(stem.vertAtt, dirVec.to_track_quat('Z', 'Y'), stem.segMax)
    upRotMat = Matrix.Rotation(-curveUpAng, 3, upRotAxis)
    dirVec.rotate(upRotMat)

    dirVec.normalize()
    dirVec *= stem.segL * tf

    # Get the end point position
    end_co = stem.p.co.copy()

    stem.spline.bezier_points.add(1)
    newPoint = stem.spline.bezier_points[-1]
    (newPoint.co, newPoint.handle_left_type, newPoint.handle_right_type) = (end_co + dirVec, hType, hType)
    newPoint.radius = stem.radS * (1 - (stem.seg + 1) / stem.segMax) + \
                      stem.radE * ((stem.seg + 1) / stem.segMax)

    if (stem.seg == stem.segMax - 1) and closeTip:
        newPoint.radius = 0.0
    # There are some cases where a point cannot have handles as VECTOR straight away, set these now
    if len(stem.spline.bezier_points) == 2:
        tempPoint = stem.spline.bezier_points[0]
        (tempPoint.handle_left_type, tempPoint.handle_right_type) = ('VECTOR', 'VECTOR')
    # Update the last point in the spline to be the newly added one
    stem.updateEnd()
    # return splineList


def genLeafMesh(leafScale, leafScaleX, leafScaleT, leafScaleV, loc, quat,
                offset, index, downAngle, downAngleV, rotate, rotateV, oldRot,
                bend, leaves, leafShape, leafangle, horzLeaves):
    if leafShape == 'hex':
        verts = [
            Vector((0, 0, 0)), Vector((0.5, 0, 1 / 3)), Vector((0.5, 0, 2 / 3)),
            Vector((0, 0, 1)), Vector((-0.5, 0, 2 / 3)), Vector((-0.5, 0, 1 / 3))
            ]
        edges = [[0, 1], [1, 2], [2, 3], [3, 4], [4, 5], [5, 0], [0, 3]]
        faces = [[0, 1, 2, 3], [0, 3, 4, 5]]
    elif leafShape == 'rect':
        # verts = [Vector((1, 0, 0)), Vector((1, 0, 1)), Vector((-1, 0, 1)), Vector((-1, 0, 0))]
        verts = [Vector((.5, 0, 0)), Vector((.5, 0, 1)), Vector((-.5, 0, 1)), Vector((-.5, 0, 0))]
        edges = [[0, 1], [1, 2], [2, 3], [3, 0]]
        faces = [[0, 1, 2, 3]]
    elif leafShape == 'dFace':
        verts = [Vector((.5, .5, 0)), Vector((.5, -.5, 0)), Vector((-.5, -.5, 0)), Vector((-.5, .5, 0))]
        edges = [[0, 1], [1, 2], [2, 3], [3, 0]]
        faces = [[0, 3, 2, 1]]
    elif leafShape == 'dVert':
        verts = [Vector((0, 0, 1))]
        edges = []
        faces = []

    vertsList = []
    facesList = []
    normal = Vector((0, 0, 1))

    if leaves < 0:
        rotMat = Matrix.Rotation(oldRot, 3, 'Y')
    else:
        rotMat = Matrix.Rotation(oldRot, 3, 'Z')

    # If the -ve flag for rotate is used we need to find which side of the stem
    # the last child point was and then grow in the opposite direction
    if rotate < 0.0:
        oldRot = -copysign(rotate + uniform(-rotateV, rotateV), oldRot)
    else:
        # If the special -ve flag for leaves is used we need a different rotation of the leaf geometry
        if leaves == -1:
            # oldRot = 0
            rotMat = Matrix.Rotation(0, 3, 'Y')
        elif leaves < -1:
            oldRot += rotate / (-leaves - 1)
        else:
            oldRot += rotate + uniform(-rotateV, rotateV)
    """
    if leaves < 0:
        rotMat = Matrix.Rotation(oldRot, 3, 'Y')
    else:
        rotMat = Matrix.Rotation(oldRot, 3, 'Z')
    """
    if leaves >= 0:
        # downRotMat = Matrix.Rotation(downAngle+uniform(-downAngleV, downAngleV), 3, 'X')

        if downAngleV > 0.0:
            downV = -downAngleV * offset
        else:
            downV = uniform(-downAngleV, downAngleV)
        downRotMat = Matrix.Rotation(downAngle + downV, 3, 'X')

    # leaf scale variation
    if (leaves < -1) and (rotate != 0):
        f = 1 - abs((oldRot - (rotate / (-leaves - 1))) / (rotate / 2))
    else:
        f = offset

    if leafScaleT < 0:
        leafScale = leafScale * (1 - (1 - f) * -leafScaleT)
    else:
        leafScale = leafScale * (1 - f * leafScaleT)

    leafScale = leafScale * uniform(1 - leafScaleV, 1 + leafScaleV)

    if leafShape == 'dFace':
        leafScale = leafScale * .1

    # If the bending of the leaves is used we need to rotate them differently
    if (bend != 0.0) and (leaves >= 0):
        normal = yAxis.copy()
        orientationVec = zAxis.copy()

        normal.rotate(quat)
        orientationVec.rotate(quat)

        thetaPos = atan2(loc.y, loc.x)
        thetaBend = thetaPos - atan2(normal.y, normal.x)
        rotateZ = Matrix.Rotation(bend * thetaBend, 3, 'Z')
        normal.rotate(rotateZ)
        orientationVec.rotate(rotateZ)

        phiBend = atan2((normal.xy).length, normal.z)
        orientation = atan2(orientationVec.y, orientationVec.x)
        rotateZOrien = Matrix.Rotation(orientation, 3, 'X')

        rotateX = Matrix.Rotation(bend * phiBend, 3, 'Z')

        rotateZOrien2 = Matrix.Rotation(-orientation, 3, 'X')

    # For each of the verts we now rotate and scale them, then append them to the list to be added to the mesh
    for v in verts:
        v.z *= leafScale
        v.y *= leafScale
        v.x *= leafScaleX * leafScale

        v.rotate(Euler((0, 0, radians(180))))

        # leafangle
        v.rotate(Matrix.Rotation(radians(-leafangle), 3, 'X'))

        if rotate < 0:
            v.rotate(Euler((0, 0, radians(90))))
            if oldRot < 0:
                v.rotate(Euler((0, 0, radians(180))))

        if (leaves > 0) and (rotate > 0) and horzLeaves:
            nRotMat = Matrix.Rotation(-oldRot + rotate, 3, 'Z')
            v.rotate(nRotMat)

        if leaves > 0:
            v.rotate(downRotMat)

        v.rotate(rotMat)
        v.rotate(quat)

        if (bend != 0.0) and (leaves > 0):
            # Correct the rotation
            v.rotate(rotateZ)
            v.rotate(rotateZOrien)
            v.rotate(rotateX)
            v.rotate(rotateZOrien2)

    if leafShape == 'dVert':
        normal = verts[0]
        normal.normalize()
        v = loc
        vertsList.append([v.x, v.y, v.z])
    else:
        for v in verts:
            v += loc
            vertsList.append([v.x, v.y, v.z])
        for f in faces:
            facesList.append([f[0] + index, f[1] + index, f[2] + index, f[3] + index])

    return vertsList, facesList, normal, oldRot


def create_armature(armAnim, leafP, cu, frameRate, leafMesh, leafObj, leafVertSize, leaves,
                    levelCount, splineToBone, treeOb, wind, gust, gustF, af1, af2, af3,
                    leafAnim, loopFrames, previewArm, armLevels, makeMesh, boneStep):
    arm = bpy.data.armatures.new('tree')
    armOb = bpy.data.objects.new('treeArm', arm)
    bpy.context.scene.collection.objects.link(armOb)
    # Create a new action to store all animation
    newAction = bpy.data.actions.new(name='windAction')
    armOb.animation_data_create()
    armOb.animation_data.action = newAction
    arm.display_type = 'STICK'
    # Add the armature modifier to the curve
    armMod = treeOb.modifiers.new('windSway', 'ARMATURE')
    if previewArm:
        armMod.show_viewport = False
        arm.display_type = 'WIRE'
        treeOb.hide_viewport = True
    armMod.use_apply_on_spline = True
    armMod.object = armOb
    armMod.use_bone_envelopes = True
    armMod.use_vertex_groups = False  # curves don't have vertex groups (yet)
    # If there are leaves then they need a modifier
    if leaves:
        armMod = leafObj.modifiers.new('windSway', 'ARMATURE')
        armMod.object = armOb
        armMod.use_bone_envelopes = False
        armMod.use_vertex_groups = True

    # Make sure all objects are deselected (may not be required?)
    for ob in bpy.context.view_layer.objects:
        ob.select_set(state=False)

    fps = bpy.context.scene.render.fps
    animSpeed = (24 / fps) * frameRate

    # Set the armature as active and go to edit mode to add bones
    bpy.context.view_layer.objects.active = armOb
    bpy.ops.object.mode_set(mode='EDIT')
    # For all the splines in the curve we need to add bones at each bezier point
    for i, parBone in enumerate(splineToBone):
        if (i < levelCount[armLevels]) or (armLevels == -1) or (not makeMesh):
            s = cu.splines[i]
            b = None
            # Get some data about the spline like length and number of points
            numPoints = len(s.bezier_points) - 1

            # find branching level
            level = 0
            for l, c in enumerate(levelCount):
                if i < c:
                    level = l
                    break
            level = min(level, 3)

            step = boneStep[level]

            # Calculate things for animation
            if armAnim:
                splineL = numPoints * ((s.bezier_points[0].co - s.bezier_points[1].co).length)
                # Set the random phase difference of the animation
                bxOffset = uniform(0, tau)
                byOffset = uniform(0, tau)
                # Set the phase multiplier for the spline
                # bMult_r = (s.bezier_points[0].radius / max(splineL, 1e-6)) * (1 / 15) * (1 / frameRate)
                # This shouldn't have to be in degrees but it looks much better in animation
                # bMult = degrees(bMult_r)
                bMult = (1 / max(splineL ** .5, 1e-6)) * (1 / 4)
                # print((1 / bMult) * tau) #print wavelength in frames

                windFreq1 = bMult * animSpeed
                windFreq2 = 0.7 * bMult * animSpeed
                if loopFrames != 0:
                    bMult_l = 1 / (loopFrames / tau)
                    fRatio = max(1, round(windFreq1 / bMult_l))
                    fgRatio = max(1, round(windFreq2 / bMult_l))
                    windFreq1 = fRatio * bMult_l
                    windFreq2 = fgRatio * bMult_l

            # For all the points in the curve (less the last) add a bone and name it by the spline it will affect
            nx = 0
            for n in range(0, numPoints, step):
                oldBone = b
                boneName = 'bone' + (str(i)).rjust(3, '0') + '.' + (str(n)).rjust(3, '0')
                b = arm.edit_bones.new(boneName)
                b.head = s.bezier_points[n].co
                nx += step
                nx = min(nx, numPoints)
                b.tail = s.bezier_points[nx].co

                b.head_radius = s.bezier_points[n].radius
                b.tail_radius = s.bezier_points[n + 1].radius
                b.envelope_distance = 0.001
                """
                # If there are leaves then we need a new vertex group so they will attach to the bone
                if not leafAnim:
                    if (len(levelCount) > 1) and (i >= levelCount[-2]) and leafObj:
                        leafObj.vertex_groups.new(name=boneName)
                    elif (len(levelCount) == 1) and leafObj:
                        leafObj.vertex_groups.new(name=boneName)
                """
                # If this is first point of the spline then it must be parented to the level above it
                if n == 0:
                    if parBone:
                        b.parent = arm.edit_bones[parBone]
                # Otherwise, we need to attach it to the previous bone in the spline
                else:
                    b.parent = oldBone
                    b.use_connect = True
                # If there isn't a previous bone then it shouldn't be attached
                if not oldBone:
                    b.use_connect = False

                # Add the animation to the armature if required
                if armAnim:
                    # Define all the required parameters of the wind sway by the dimension of the spline
                    # a0 = 4 * splineL * (1 - n / (numPoints + 1)) / max(s.bezier_points[n].radius, 1e-6)
                    a0 = 2 * (splineL / numPoints) * (1 - n / (numPoints + 1)) / max(s.bezier_points[n].radius, 1e-6)
                    a0 = a0 * min(step, numPoints)
                    # a0 = (splineL / numPoints) / max(s.bezier_points[n].radius, 1e-6)
                    a1 = (wind / 50) * a0
                    a2 = a1 * .65  # (windGust / 50) * a0 + a1 / 2

                    p = s.bezier_points[nx].co - s.bezier_points[n].co
                    p.normalize()
                    ag = (wind * gust / 50) * a0
                    a3 = -p[0] * ag
                    a4 = p[2] * ag

                    a1 = radians(a1)
                    a2 = radians(a2)
                    a3 = radians(a3)
                    a4 = radians(a4)

                    # wind bending
                    if loopFrames == 0:
                        swayFreq = gustF * (tau / fps) * frameRate  # animSpeed # .075 # 0.02
                    else:
                        swayFreq = 1 / (loopFrames / tau)

                    # Prevent tree base from rotating
                    if (boneName == "bone000.000") or (boneName == "bone000.001"):
                        a1 = 0
                        a2 = 0
                        a3 = 0
                        a4 = 0

                    # Add new fcurves for each sway as well as the modifiers
                    swayX = armOb.animation_data.action.fcurves.new(
                                            'pose.bones["' + boneName + '"].rotation_euler', index=0
                                            )
                    swayY = armOb.animation_data.action.fcurves.new(
                                            'pose.bones["' + boneName + '"].rotation_euler', index=2
                                            )
                    swayXMod1 = swayX.modifiers.new(type='FNGENERATOR')
                    swayXMod2 = swayX.modifiers.new(type='FNGENERATOR')

                    swayYMod1 = swayY.modifiers.new(type='FNGENERATOR')
                    swayYMod2 = swayY.modifiers.new(type='FNGENERATOR')

                    # Set the parameters for each modifier
                    swayXMod1.amplitude = a1
                    swayXMod1.phase_offset = bxOffset
                    swayXMod1.phase_multiplier = windFreq1

                    swayXMod2.amplitude = a2
                    swayXMod2.phase_offset = 0.7 * bxOffset
                    swayXMod2.phase_multiplier = windFreq2
                    swayXMod2.use_additive = True

                    swayYMod1.amplitude = a1
                    swayYMod1.phase_offset = byOffset
                    swayYMod1.phase_multiplier = windFreq1

                    swayYMod2.amplitude = a2
                    swayYMod2.phase_offset = 0.7 * byOffset
                    swayYMod2.phase_multiplier = windFreq2
                    swayYMod2.use_additive = True

                    # wind bending
                    swayYMod3 = swayY.modifiers.new(type='FNGENERATOR')
                    swayYMod3.amplitude = a3
                    swayYMod3.phase_multiplier = swayFreq
                    swayYMod3.value_offset = .6 * a3
                    swayYMod3.use_additive = True

                    swayXMod3 = swayX.modifiers.new(type='FNGENERATOR')
                    swayXMod3.amplitude = a4
                    swayXMod3.phase_multiplier = swayFreq
                    swayXMod3.value_offset = .6 * a4
                    swayXMod3.use_additive = True

    if leaves:
        bonelist = [b.name for b in arm.edit_bones]
        vertexGroups = OrderedDict()
        for i, cp in enumerate(leafP):
            # find leafs parent bone
            leafParent = roundBone(cp.parBone, boneStep[armLevels])
            idx = int(leafParent[4:-4])
            while leafParent not in bonelist:
                # find parent bone of parent bone
                leafParent = splineToBone[idx]
                idx = int(leafParent[4:-4])

            if leafAnim:
                bname = "leaf" + str(i)
                b = arm.edit_bones.new(bname)
                b.head = cp.co
                b.tail = cp.co + Vector((0, 0, .02))
                b.envelope_distance = 0.0
                b.parent = arm.edit_bones[leafParent]

                vertexGroups[bname] = [
                                    v.index for v in
                                    leafMesh.vertices[leafVertSize * i:(leafVertSize * i + leafVertSize)]
                                    ]

                if armAnim:
                    # Define all the required parameters of the wind sway by the dimension of the spline
                    a1 = wind * .25
                    a1 *= af1

                    bMult = (1 / animSpeed) * 6
                    bMult *= 1 / max(af2, .001)

                    ofstRand = af3
                    bxOffset = uniform(-ofstRand, ofstRand)
                    byOffset = uniform(-ofstRand, ofstRand)

                    # Add new fcurves for each sway as well as the modifiers
                    swayX = armOb.animation_data.action.fcurves.new(
                                                'pose.bones["' + bname + '"].rotation_euler', index=0
                                                )
                    swayY = armOb.animation_data.action.fcurves.new(
                                                'pose.bones["' + bname + '"].rotation_euler', index=2
                                                )
                    # Add keyframe so noise works
                    swayX.keyframe_points.add(1)
                    swayY.keyframe_points.add(1)
                    swayX.keyframe_points[0].co = (0, 0)
                    swayY.keyframe_points[0].co = (0, 0)

                    # Add noise modifiers
                    swayXMod = swayX.modifiers.new(type='NOISE')
                    swayYMod = swayY.modifiers.new(type='NOISE')

                    if loopFrames != 0:
                        swayXMod.use_restricted_range = True
                        swayXMod.frame_end = loopFrames
                        swayXMod.blend_in = 4
                        swayXMod.blend_out = 4
                        swayYMod.use_restricted_range = True
                        swayYMod.frame_end = loopFrames
                        swayYMod.blend_in = 4
                        swayYMod.blend_out = 4

                    swayXMod.scale = bMult
                    swayXMod.strength = a1
                    swayXMod.offset = bxOffset

                    swayYMod.scale = bMult
                    swayYMod.strength = a1
                    swayYMod.offset = byOffset

            else:
                if leafParent not in vertexGroups:
                    vertexGroups[leafParent] = []
                vertexGroups[leafParent].extend(
                                        [v.index for v in
                                        leafMesh.vertices[leafVertSize * i:(leafVertSize * i + leafVertSize)]]
                                        )

        for group in vertexGroups:
            leafObj.vertex_groups.new(name=group)
            leafObj.vertex_groups[group].add(vertexGroups[group], 1.0, 'ADD')

    # Now we need the rotation mode to be 'XYZ' to ensure correct rotation
    bpy.ops.object.mode_set(mode='OBJECT')
    for p in armOb.pose.bones:
        p.rotation_mode = 'XYZ'
    treeOb.parent = armOb


def kickstart_trunk(addstem, levels, leaves, branches, cu, curve, curveRes,
                    curveV, attractUp, length, lengthV, ratio, ratioPower,
                    resU, scale0, scaleV0, scaleVal, taper, minRadius, rootFlare):
    newSpline = cu.splines.new('BEZIER')
    cu.resolution_u = resU
    newPoint = newSpline.bezier_points[-1]
    newPoint.co = Vector((0, 0, 0))
    newPoint.handle_right = Vector((0, 0, 1))
    newPoint.handle_left = Vector((0, 0, -1))
    # (newPoint.handle_right_type, newPoint.handle_left_type) = ('VECTOR', 'VECTOR')
    branchL = scaleVal * length[0]
    curveVal = curve[0] / curveRes[0]
    # curveVal = curveVal * (branchL / scaleVal)
    if levels == 1:
        childStems = leaves
    else:
        childStems = branches[1]
    startRad = scaleVal * ratio * scale0 * uniform(1 - scaleV0, 1 + scaleV0)  # * (scale0 + uniform(-scaleV0, scaleV0))
    endRad = (startRad * (1 - taper[0])) ** ratioPower
    startRad = max(startRad, minRadius)
    endRad = max(endRad, minRadius)
    newPoint.radius = startRad * rootFlare
    addstem(
        stemSpline(
                newSpline, curveVal, curveV[0] / curveRes[0], attractUp[0],
                0, curveRes[0], branchL / curveRes[0],
                childStems, startRad, endRad, 0, 0, None
                )
            )


def fabricate_stems(addsplinetobone, addstem, baseSize, branches, childP, cu, curve, curveBack,
                    curveRes, curveV, attractUp, downAngle, downAngleV, leafDist, leaves, length,
                    lengthV, levels, n, ratioPower, resU, rotate, rotateV, scaleVal, shape, storeN,
                    taper, shapeS, minRadius, radiusTweak, customShape, rMode, segSplits,
                    useOldDownAngle, useParentAngle, boneStep):

    # prevent baseSize from going to 1.0
    baseSize = min(0.999, baseSize)

    # Store the old rotation to allow new stems to be rotated away from the previous one.
    oldRotate = 0

    # use fancy child point selection / rotation
    if (n == 1) and (rMode != "original"):
        childP_T = OrderedDict()
        childP_L = []
        for p in childP:
            if p.offset == 1:
                childP_L.append(p)
            else:
                if p.offset not in childP_T:
                    childP_T[p.offset] = [p]
                else:
                    childP_T[p.offset].append(p)

        childP_T = [childP_T[k] for k in sorted(childP_T.keys())]

        childP = []
        rot_a = []
        for p in childP_T:
            if rMode == "rotate":
                if rotate[n] < 0.0:
                    oldRotate = -copysign(rotate[n], oldRotate)
                else:
                    oldRotate += rotate[n]
                bRotate = oldRotate + uniform(-rotateV[n], rotateV[n])

                # choose start point whose angle is closest to the rotate angle
                a1 = bRotate % tau
                a_diff = []
                for a in p:
                    a2 = atan2(a.co[0], -a.co[1])
                    d = min((a1 - a2 + tau) % tau, (a2 - a1 + tau) % tau)
                    a_diff.append(d)

                idx = a_diff.index(min(a_diff))

                # find actual rotate angle from branch location
                br = p[idx]
                b = br.co
                vx = sin(bRotate)
                vy = cos(bRotate)
                v = Vector((vx, vy))

                bD = ((b[0] * b[0] + b[1] * b[1]) ** .5)
                bL = br.lengthPar * length[1] * shapeRatio(shape, (1 - br.offset) / (1 - baseSize), custom=customShape)

                # account for down angle
                if downAngleV[1] > 0:
                    downA = downAngle[n] + (-downAngleV[n] * (1 - (1 - br.offset) / (1 - baseSize)) ** 2)
                else:
                    downA = downAngle[n]
                if downA < (.5 * pi):
                    downA = sin(downA) ** 2
                    bL *= downA

                bL *= 0.33
                v *= (bD + bL)

                bv = Vector((b[0], -b[1]))
                cv = v - bv
                a = atan2(cv[0], cv[1])
                # rot_a.append(a)
                """
                # add fill points at top  #experimental
                fillHeight = 1 - degrees(rotateV[3]) # 0.8
                if fillHeight < 1:
                    w = (p[0].offset - fillHeight) / (1- fillHeight)
                    prob_b = random() < w
                else:
                    prob_b = False

                if (p[0].offset > fillHeight): # prob_b and (len(p) > 1):  ##(p[0].offset > fillHeight) and
                    childP.append(p[randint(0, len(p)-1)])
                    rot_a.append(bRotate)# + pi)
                """
                childP.append(p[idx])
                rot_a.append(a)

            else:
                idx = randint(0, len(p) - 1)
                childP.append(p[idx])
            # childP.append(p[idx])

        childP.extend(childP_L)
        rot_a.extend([0] * len(childP_L))

        oldRotate = 0

    for i, p in enumerate(childP):
        # Add a spline and set the coordinate of the first point.
        newSpline = cu.splines.new('BEZIER')
        cu.resolution_u = resU
        newPoint = newSpline.bezier_points[-1]
        newPoint.co = p.co
        tempPos = zAxis.copy()
        # If the -ve flag for downAngle is used we need a special formula to find it
        if useOldDownAngle:
            if downAngleV[n] < 0.0:
                downV = downAngleV[n] * (1 - 2 * (.2 + .8 * ((1 - p.offset) / (1 - baseSize))))
            # Otherwise just find a random value
            else:
                downV = uniform(-downAngleV[n], downAngleV[n])
        else:
            if downAngleV[n] < 0.0:
                downV = uniform(-downAngleV[n], downAngleV[n])
            else:
                downV = -downAngleV[n] * (1 - (1 - p.offset) / (1 - baseSize)) ** 2  # (110, 80) = (60, -50)

        if p.offset == 1:
            downRotMat = Matrix.Rotation(0, 3, 'X')
        else:
            downRotMat = Matrix.Rotation(downAngle[n] + downV, 3, 'X')

        # If the -ve flag for rotate is used we need to find which side of the stem
        # the last child point was and then grow in the opposite direction
        if rotate[n] < 0.0:
            oldRotate = -copysign(rotate[n], oldRotate)
        # Otherwise just generate a random number in the specified range
        else:
            oldRotate += rotate[n]
        bRotate = oldRotate + uniform(-rotateV[n], rotateV[n])

        if (n == 1) and (rMode == "rotate"):
            bRotate = rot_a[i]

        rotMat = Matrix.Rotation(bRotate, 3, 'Z')

        # Rotate the direction of growth and set the new point coordinates
        tempPos.rotate(downRotMat)
        tempPos.rotate(rotMat)

        # use quat angle
        if (rMode == "rotate") and (n == 1) and (p.offset != 1):
            if useParentAngle:
                edir = p.quat.to_euler('XYZ', Euler((0, 0, bRotate), 'XYZ'))
                edir[0] = 0
                edir[1] = 0

                edir[2] = -edir[2]
                tempPos.rotate(edir)

                dec = declination(p.quat)
                tempPos.rotate(Matrix.Rotation(radians(dec), 3, 'X'))

                edir[2] = -edir[2]
                tempPos.rotate(edir)
        else:
            tempPos.rotate(p.quat)

        newPoint.handle_right = p.co + tempPos

        # Make length variation inversely proportional to segSplits
        # lenV = (1 - min(segSplits[n], 1)) * lengthV[n]

        # Find branch length and the number of child stems.
        maxbL = scaleVal
        for l in length[:n + 1]:
            maxbL *= l
        lMax = length[n]  # * uniform(1 - lenV, 1 + lenV)
        if n == 1:
            lShape = shapeRatio(shape, (1 - p.stemOffset) / (1 - baseSize), custom=customShape)
        else:
            lShape = shapeRatio(shapeS, (1 - p.stemOffset) / (1 - baseSize))
        branchL = p.lengthPar * lMax * lShape
        childStems = branches[min(3, n + 1)] * (0.1 + 0.9 * (branchL / maxbL))

        # If this is the last level before leaves then we need to generate the child points differently
        if (storeN == levels - 1):
            if leaves < 0:
                childStems = False
            else:
                childStems = leaves * (0.1 + 0.9 * (branchL / maxbL)) * shapeRatio(leafDist, (1 - p.offset))

        # print("n=%d, levels=%d, n'=%d, childStems=%s"%(n, levels, storeN, childStems))

        # Determine the starting and ending radii of the stem using the tapering of the stem
        startRad = min((p.radiusPar[0] * ((branchL / p.lengthPar) ** ratioPower)) * radiusTweak[n], p.radiusPar[1])
        if p.offset == 1:
            startRad = p.radiusPar[1]
        endRad = (startRad * (1 - taper[n])) ** ratioPower
        startRad = max(startRad, minRadius)
        endRad = max(endRad, minRadius)
        newPoint.radius = startRad

        # stem curvature
        curveVal = curve[n] / curveRes[n]
        curveVar = curveV[n] / curveRes[n]

        # curveVal = curveVal * (branchL / scaleVal)

        # Add the new stem to list of stems to grow and define which bone it will be parented to
        addstem(
            stemSpline(
                newSpline, curveVal, curveVar, attractUp[n],
                0, curveRes[n], branchL / curveRes[n], childStems,
                startRad, endRad, len(cu.splines) - 1, 0, p.quat
                )
            )

        bone = roundBone(p.parBone, boneStep[n - 1])
        if p.offset == 1:
            isend = True
        else:
            isend = False
        addsplinetobone((bone, isend))


def perform_pruning(baseSize, baseSplits, childP, cu, currentMax, currentMin, currentScale, curve,
                    curveBack, curveRes, deleteSpline, forceSprout, handles, n, oldMax, originalSplineToBone,
                    originalCo, originalCurv, originalCurvV, originalHandleL, originalHandleR, originalLength,
                    originalSeg, prune, prunePowerHigh, prunePowerLow, pruneRatio, pruneWidth, pruneBase,
                    pruneWidthPeak, randState, ratio, scaleVal, segSplits, splineToBone, splitAngle, splitAngleV,
                    st, startPrune, branchDist, length, splitByLen, closeTip, nrings, splitBias, splitHeight,
                    attractOut, rMode, lengthV, taperCrown, boneStep, rotate, rotateV):
    while startPrune and ((currentMax - currentMin) > 0.005):
        setstate(randState)

        # If the search will halt after this iteration, then set the adjustment of stem
        # length to take into account the pruning ratio
        if (currentMax - currentMin) < 0.01:
            currentScale = (currentScale - 1) * pruneRatio + 1
            startPrune = False
            forceSprout = True
        # Change the segment length of the stem by applying some scaling
        st.segL = originalLength * currentScale
        # To prevent millions of splines being created we delete any old ones and
        # replace them with only their first points to begin the spline again
        if deleteSpline:
            for x in splineList:
                cu.splines.remove(x.spline)
            newSpline = cu.splines.new('BEZIER')
            newPoint = newSpline.bezier_points[-1]
            newPoint.co = originalCo
            newPoint.handle_right = originalHandleR
            newPoint.handle_left = originalHandleL
            (newPoint.handle_left_type, newPoint.handle_right_type) = ('VECTOR', 'VECTOR')
            st.spline = newSpline
            st.curv = originalCurv
            st.curvV = originalCurvV
            st.seg = originalSeg
            st.p = newPoint
            newPoint.radius = st.radS
            splineToBone = originalSplineToBone

        # Initialise the spline list for those contained in the current level of branching
        splineList = [st]

        # split length variation
        stemsegL = splineList[0].segL  # initial segment length used for variation
        splineList[0].segL = stemsegL * uniform(1 - lengthV[n], 1 + lengthV[n])  # variation for first stem

        # For each of the segments of the stem which must be grown we have to add to each spline in splineList
        for k in range(curveRes[n]):
            # Make a copy of the current list to avoid continually adding to the list we're iterating over
            tempList = splineList[:]
            # print('Leng: ', len(tempList))

            # for curve variation
            if curveRes[n] > 1:
                kp = (k / (curveRes[n] - 1))  # * 2
            else:
                kp = 1.0

            # split bias
            splitValue = segSplits[n]
            if n == 0:
                splitValue = ((2 * splitBias) * (kp - .5) + 1) * splitValue
                splitValue = max(splitValue, 0.0)

            # For each of the splines in this list set the number of splits and then grow it
            for spl in tempList:
                # adjust numSplit
                lastsplit = getattr(spl, 'splitlast', 0)
                splitVal = splitValue
                if lastsplit == 0:
                    splitVal = splitValue * 1.33
                elif lastsplit == 1:
                    splitVal = splitValue * splitValue

                if k == 0:
                    numSplit = 0
                elif (n == 0) and (k < ((curveRes[n] - 1) * splitHeight)) and (k != 1):
                    numSplit = 0
                elif (k == 1) and (n == 0):
                    numSplit = baseSplits
                # always split at splitHeight
                elif (n == 0) and (k == int((curveRes[n] - 1) * splitHeight) + 1) and (splitVal > 0):
                    numSplit = 1
                else:
                    if (n >= 1) and splitByLen:
                        L = ((spl.segL * curveRes[n]) / scaleVal)
                        lf = 1
                        for l in length[:n + 1]:
                            lf *= l
                        L = L / lf
                        numSplit = splits2(splitVal * L)
                    else:
                        numSplit = splits2(splitVal)

                if (k == int(curveRes[n] / 2 + 0.5)) and (curveBack[n] != 0):
                    spl.curv += 2 * (curveBack[n] / curveRes[n])  # was -4 *

                growSpline(
                        n, spl, numSplit, splitAngle[n], splitAngleV[n], splineList,
                        handles, splineToBone, closeTip, kp, splitHeight, attractOut[n],
                        stemsegL, lengthV[n], taperCrown, boneStep, rotate, rotateV
                        )

        # If pruning is enabled then we must check to see if the end of the spline is within the envelope
        if prune:
            # Check each endpoint to see if it is inside
            for s in splineList:
                coordMag = (s.spline.bezier_points[-1].co.xy).length
                ratio = (scaleVal - s.spline.bezier_points[-1].co.z) / (scaleVal * max(1 - pruneBase, 1e-6))
                # Don't think this if part is needed
                if (n == 0) and (s.spline.bezier_points[-1].co.z < pruneBase * scaleVal):
                    insideBool = True  # Init to avoid UnboundLocalError later
                else:
                    insideBool = (
                    (coordMag / scaleVal) < pruneWidth * shapeRatio(9, ratio, pruneWidthPeak, prunePowerHigh,
                                                                    prunePowerLow))
                # If the point is not inside then we adjust the scale and current search bounds
                if not insideBool:
                    oldMax = currentMax
                    currentMax = currentScale
                    currentScale = 0.5 * (currentMax + currentMin)
                    break
            # If the scale is the original size and the point is inside then
            # we need to make sure it won't be pruned or extended to the edge of the envelope
            if insideBool and (currentScale != 1):
                currentMin = currentScale
                currentMax = oldMax
                currentScale = 0.5 * (currentMax + currentMin)
            if insideBool and ((currentMax - currentMin) == 1):
                currentMin = 1

        # If the search will halt on the next iteration then we need
        # to make sure we sprout child points to grow the next splines or leaves
        if (((currentMax - currentMin) < 0.005) or not prune) or forceSprout:
            if (n == 0) and (rMode != "original"):
                tVals = findChildPoints2(splineList, st.children)
            else:
                tVals = findChildPoints(splineList, st.children)
            # print("debug tvals[%d] , splineList[%d], %s" % ( len(tVals), len(splineList), st.children))
            # If leaves is -ve then we need to make sure the only point which sprouts is the end of the spline
            if not st.children:
                tVals = [1.0]
            # remove some of the points because of baseSize
            trimNum = int(baseSize * (len(tVals) + 1))
            tVals = tVals[trimNum:]

            # grow branches in rings
            if (n == 0) and (nrings > 0):
                # tVals = [(floor(t * nrings)) / nrings for t in tVals[:-1]]
                tVals = [(floor(t * nrings) / nrings) * uniform(.995, 1.005) for t in tVals[:-1]]
                tVals.append(1)
                tVals = [t for t in tVals if t > baseSize]

            # branch distribution
            if n == 0:
                tVals = [((t - baseSize) / (1 - baseSize)) for t in tVals]
                if branchDist < 1.0:
                    tVals = [t ** (1 / branchDist) for t in tVals]
                else:
                    tVals = [1 - (1 - t) ** branchDist for t in tVals]
                tVals = [t * (1 - baseSize) + baseSize for t in tVals]

            # For all the splines, we interpolate them and add the new points to the list of child points
            maxOffset = max([s.offsetLen + (len(s.spline.bezier_points) - 1) * s.segL for s in splineList])
            for s in splineList:
                # print(str(n)+'level: ', s.segMax*s.segL)
                childP.extend(interpStem(s, tVals, s.segMax * s.segL, s.radS, maxOffset, baseSize))

        # Force the splines to be deleted
        deleteSpline = True
        # If pruning isn't enabled then make sure it doesn't loop
        if not prune:
            startPrune = False
    return ratio, splineToBone


# calculate taper automatically
def findtaper(length, taper, shape, shapeS, levels, customShape):
    taperS = []
    for i, t in enumerate(length):
        if i == 0:
            shp = 1.0
        elif i == 1:
            shp = shapeRatio(shape, 0, custom=customShape)
        else:
            shp = shapeRatio(shapeS, 0)
        t = t * shp
        taperS.append(t)

    taperP = []
    for i, t in enumerate(taperS):
        pm = 1
        for x in range(i + 1):
            pm *= taperS[x]
        taperP.append(pm)

    taperR = []
    for i, t in enumerate(taperP):
        t = sum(taperP[i:levels])
        taperR.append(t)

    taperT = []
    for i, t in enumerate(taperR):
        try:
            t = taperP[i] / taperR[i]
        except ZeroDivisionError:
            t = 1.0
        taperT.append(t)

    taperT = [t * taper[i] for i, t in enumerate(taperT)]

    return taperT


def addTree(props):
    global splitError
    # startTime = time.time()
    # Set the seed for repeatable results
    seed(props.seed)

    # Set all other variables
    levels = props.levels
    length = props.length
    lengthV = props.lengthV
    taperCrown = props.taperCrown
    branches = props.branches
    curveRes = props.curveRes
    curve = toRad(props.curve)
    curveV = toRad(props.curveV)
    curveBack = toRad(props.curveBack)
    baseSplits = props.baseSplits
    segSplits = props.segSplits
    splitByLen = props.splitByLen
    rMode = props.rMode
    splitAngle = toRad(props.splitAngle)
    splitAngleV = toRad(props.splitAngleV)
    scale = props.scale
    scaleV = props.scaleV
    attractUp = props.attractUp
    attractOut = props.attractOut
    shape = int(props.shape)
    shapeS = int(props.shapeS)
    customShape = props.customShape
    branchDist = props.branchDist
    nrings = props.nrings
    baseSize = props.baseSize
    baseSize_s = props.baseSize_s
    splitHeight = props.splitHeight
    splitBias = props.splitBias
    ratio = props.ratio
    minRadius = props.minRadius
    closeTip = props.closeTip
    rootFlare = props.rootFlare
    autoTaper = props.autoTaper
    taper = props.taper
    radiusTweak = props.radiusTweak
    ratioPower = props.ratioPower
    downAngle = toRad(props.downAngle)
    downAngleV = toRad(props.downAngleV)
    rotate = toRad(props.rotate)
    rotateV = toRad(props.rotateV)
    scale0 = props.scale0
    scaleV0 = props.scaleV0
    prune = props.prune
    pruneWidth = props.pruneWidth
    pruneBase = props.pruneBase
    pruneWidthPeak = props.pruneWidthPeak
    prunePowerLow = props.prunePowerLow
    prunePowerHigh = props.prunePowerHigh
    pruneRatio = props.pruneRatio
    leafDownAngle = radians(props.leafDownAngle)
    leafDownAngleV = radians(props.leafDownAngleV)
    leafRotate = radians(props.leafRotate)
    leafRotateV = radians(props.leafRotateV)
    leafScale = props.leafScale
    leafScaleX = props.leafScaleX
    leafScaleT = props.leafScaleT
    leafScaleV = props.leafScaleV
    leafShape = props.leafShape
    leafDupliObj = props.leafDupliObj
    bend = props.bend
    leafangle = props.leafangle
    horzLeaves = props.horzLeaves
    leafDist = int(props.leafDist)
    bevelRes = props.bevelRes
    resU = props.resU

    useArm = props.useArm
    previewArm = props.previewArm
    armAnim = props.armAnim
    leafAnim = props.leafAnim
    frameRate = props.frameRate
    loopFrames = props.loopFrames

    # windSpeed = props.windSpeed
    # windGust = props.windGust

    wind = props.wind
    gust = props.gust
    gustF = props.gustF

    af1 = props.af1
    af2 = props.af2
    af3 = props.af3

    makeMesh = props.makeMesh
    armLevels = props.armLevels
    boneStep = props.boneStep

    useOldDownAngle = props.useOldDownAngle
    useParentAngle = props.useParentAngle

    if not makeMesh:
        boneStep = [1, 1, 1, 1]

    # taper
    if autoTaper:
        taper = findtaper(length, taper, shape, shapeS, levels, customShape)
        # pLevels = branches[0]
        # taper = findtaper(length, taper, shape, shapeS, pLevels, customShape)

    leafObj = None

    # Some effects can be turned ON and OFF, the necessary variables are changed here
    if not props.bevel:
        bevelDepth = 0.0
    else:
        bevelDepth = 1.0

    if not props.showLeaves:
        leaves = 0
    else:
        leaves = props.leaves

    if props.handleType == '0':
        handles = 'AUTO'
    else:
        handles = 'VECTOR'

    for ob in bpy.context.view_layer.objects:
        ob.select_set(state=False)

    # Initialise the tree object and curve and adjust the settings
    cu = bpy.data.curves.new('tree', 'CURVE')
    treeOb = bpy.data.objects.new('tree', cu)
    bpy.context.scene.collection.objects.link(treeOb)

    # treeOb.location=bpy.context.scene.cursor.location attractUp

    cu.dimensions = '3D'
    cu.fill_mode = 'FULL'
    cu.bevel_depth = bevelDepth
    cu.bevel_resolution = bevelRes

    # Fix the scale of the tree now
    scaleVal = scale + uniform(-scaleV, scaleV)
    scaleVal += copysign(1e-6, scaleVal)  # Move away from zero to avoid div by zero

    pruneBase = min(pruneBase, baseSize)
    # If pruning is turned on we need to draw the pruning envelope
    if prune:
        enHandle = 'VECTOR'
        enNum = 128
        enCu = bpy.data.curves.new('envelope', 'CURVE')
        enOb = bpy.data.objects.new('envelope', enCu)
        enOb.parent = treeOb
        bpy.context.scene.collection.objects.link(enOb)
        newSpline = enCu.splines.new('BEZIER')
        newPoint = newSpline.bezier_points[-1]
        newPoint.co = Vector((0, 0, scaleVal))
        (newPoint.handle_right_type, newPoint.handle_left_type) = (enHandle, enHandle)
        # Set the coordinates by varying the z value, envelope will be aligned to the x-axis
        for c in range(enNum):
            newSpline.bezier_points.add(1)
            newPoint = newSpline.bezier_points[-1]
            ratioVal = (c + 1) / (enNum)
            zVal = scaleVal - scaleVal * (1 - pruneBase) * ratioVal
            newPoint.co = Vector(
                            (
                            scaleVal * pruneWidth *
                            shapeRatio(9, ratioVal, pruneWidthPeak, prunePowerHigh, prunePowerLow),
                            0, zVal
                            )
                        )
            (newPoint.handle_right_type, newPoint.handle_left_type) = (enHandle, enHandle)
        newSpline = enCu.splines.new('BEZIER')
        newPoint = newSpline.bezier_points[-1]
        newPoint.co = Vector((0, 0, scaleVal))
        (newPoint.handle_right_type, newPoint.handle_left_type) = (enHandle, enHandle)
        # Create a second envelope but this time on the y-axis
        for c in range(enNum):
            newSpline.bezier_points.add(1)
            newPoint = newSpline.bezier_points[-1]
            ratioVal = (c + 1) / (enNum)
            zVal = scaleVal - scaleVal * (1 - pruneBase) * ratioVal
            newPoint.co = Vector(
                            (
                            0, scaleVal * pruneWidth *
                            shapeRatio(9, ratioVal, pruneWidthPeak, prunePowerHigh, prunePowerLow),
                            zVal
                            )
                        )
            (newPoint.handle_right_type, newPoint.handle_left_type) = (enHandle, enHandle)

    childP = []
    stemList = []

    levelCount = []
    splineToBone = deque([''])
    addsplinetobone = splineToBone.append

    # Each of the levels needed by the user we grow all the splines
    for n in range(levels):
        storeN = n
        stemList = deque()
        addstem = stemList.append
        # If n is used as an index to access parameters for the tree
        # it must be at most 3 or it will reference outside the array index
        n = min(3, n)
        splitError = 0.0

        # closeTip only on last level
        closeTipp = all([(n == levels - 1), closeTip])

        # If this is the first level of growth (the trunk) then we need some special work to begin the tree
        if n == 0:
            kickstart_trunk(addstem, levels, leaves, branches, cu, curve, curveRes,
                            curveV, attractUp, length, lengthV, ratio, ratioPower, resU,
                            scale0, scaleV0, scaleVal, taper, minRadius, rootFlare)
        # If this isn't the trunk then we may have multiple stem to initialise
        else:
            # For each of the points defined in the list of stem starting points we need to grow a stem.
            fabricate_stems(addsplinetobone, addstem, baseSize, branches, childP, cu, curve, curveBack,
                            curveRes, curveV, attractUp, downAngle, downAngleV, leafDist, leaves, length, lengthV,
                            levels, n, ratioPower, resU, rotate, rotateV, scaleVal, shape, storeN,
                            taper, shapeS, minRadius, radiusTweak, customShape, rMode, segSplits,
                            useOldDownAngle, useParentAngle, boneStep)

        # change base size for each level
        if n > 0:
            baseSize *= baseSize_s  # decrease at each level
        if (n == levels - 1):
            baseSize = 0

        childP = []
        # Now grow each of the stems in the list of those to be extended
        for st in stemList:
            # When using pruning, we need to ensure that the random effects
            # will be the same for each iteration to make sure the problem is linear
            randState = getstate()
            startPrune = True
            lengthTest = 0.0
            # Store all the original values for the stem to make sure
            # we have access after it has been modified by pruning
            originalLength = st.segL
            originalCurv = st.curv
            originalCurvV = st.curvV
            originalSeg = st.seg
            originalHandleR = st.p.handle_right.copy()
            originalHandleL = st.p.handle_left.copy()
            originalCo = st.p.co.copy()
            currentMax = 1.0
            currentMin = 0.0
            currentScale = 1.0
            oldMax = 1.0
            deleteSpline = False
            originalSplineToBone = copy.copy(splineToBone)
            forceSprout = False
            # Now do the iterative pruning, this uses a binary search and halts once the difference
            # between upper and lower bounds of the search are less than 0.005
            ratio, splineToBone = perform_pruning(
                                        baseSize, baseSplits, childP, cu, currentMax, currentMin,
                                        currentScale, curve, curveBack, curveRes, deleteSpline, forceSprout,
                                        handles, n, oldMax, originalSplineToBone, originalCo, originalCurv,
                                        originalCurvV, originalHandleL, originalHandleR, originalLength,
                                        originalSeg, prune, prunePowerHigh, prunePowerLow, pruneRatio,
                                        pruneWidth, pruneBase, pruneWidthPeak, randState, ratio, scaleVal,
                                        segSplits, splineToBone, splitAngle, splitAngleV, st, startPrune,
                                        branchDist, length, splitByLen, closeTipp, nrings, splitBias,
                                        splitHeight, attractOut, rMode, lengthV, taperCrown, boneStep,
                                        rotate, rotateV
                                        )

        levelCount.append(len(cu.splines))

    # If we need to add leaves, we do it here
    leafVerts = []
    leafFaces = []
    leafNormals = []

    leafMesh = None  # in case we aren't creating leaves, we'll still have the variable

    leafP = []
    if leaves:
        oldRot = 0.0
        n = min(3, n + 1)
        # For each of the child points we add leaves
        for cp in childP:
            # If the special flag is set then we need to add several leaves at the same location
            if leaves < 0:
                oldRot = -leafRotate / 2
                for g in range(abs(leaves)):
                    (vertTemp, faceTemp, normal, oldRot) = genLeafMesh(
                                                                leafScale, leafScaleX, leafScaleT,
                                                                leafScaleV, cp.co, cp.quat, cp.offset,
                                                                len(leafVerts), leafDownAngle, leafDownAngleV,
                                                                leafRotate, leafRotateV,
                                                                oldRot, bend, leaves, leafShape,
                                                                leafangle, horzLeaves
                                                                )
                    leafVerts.extend(vertTemp)
                    leafFaces.extend(faceTemp)
                    leafNormals.extend(normal)
                    leafP.append(cp)
            # Otherwise just add the leaves like splines
            else:
                (vertTemp, faceTemp, normal, oldRot) = genLeafMesh(
                                                            leafScale, leafScaleX, leafScaleT, leafScaleV,
                                                            cp.co, cp.quat, cp.offset, len(leafVerts),
                                                            leafDownAngle, leafDownAngleV, leafRotate,
                                                            leafRotateV, oldRot, bend, leaves, leafShape,
                                                            leafangle, horzLeaves
                                                            )
                leafVerts.extend(vertTemp)
                leafFaces.extend(faceTemp)
                leafNormals.extend(normal)
                leafP.append(cp)

        # Create the leaf mesh and object, add geometry using from_pydata,
        # edges are currently added by validating the mesh which isn't great
        leafMesh = bpy.data.meshes.new('leaves')
        leafObj = bpy.data.objects.new('leaves', leafMesh)
        bpy.context.scene.collection.objects.link(leafObj)
        leafObj.parent = treeOb
        leafMesh.from_pydata(leafVerts, (), leafFaces)

        # set vertex normals for dupliVerts
        if leafShape == 'dVert':
            leafMesh.vertices.foreach_set('normal', leafNormals)

        # enable duplication
        if leafShape == 'dFace':
            leafObj.instance_type = "FACES"
            leafObj.use_instance_faces_scale = True
            leafObj.instance_faces_scale = 10.0
            try:
                if leafDupliObj not in "NONE":
                    bpy.data.objects[leafDupliObj].parent = leafObj
            except KeyError:
                pass
        elif leafShape == 'dVert':
            leafObj.instance_type = "VERTS"
            leafObj.use_instance_vertices_rotation = True
            try:
                if leafDupliObj not in "NONE":
                    bpy.data.objects[leafDupliObj].parent = leafObj
            except KeyError:
                pass

        # add leaf UVs
        if leafShape == 'rect':
            leafMesh.uv_layers.new(name='leafUV')
            uvlayer = leafMesh.uv_layers.active.data

            u1 = .5 * (1 - leafScaleX)
            u2 = 1 - u1

            for i in range(0, len(leafFaces)):
                uvlayer[i * 4 + 0].uv = Vector((u2, 0))
                uvlayer[i * 4 + 1].uv = Vector((u2, 1))
                uvlayer[i * 4 + 2].uv = Vector((u1, 1))
                uvlayer[i * 4 + 3].uv = Vector((u1, 0))

        elif leafShape == 'hex':
            leafMesh.uv_layers.new(name='leafUV')
            uvlayer = leafMesh.uv_layers.active.data

            u1 = .5 * (1 - leafScaleX)
            u2 = 1 - u1

            for i in range(0, int(len(leafFaces) / 2)):
                uvlayer[i * 8 + 0].uv = Vector((.5, 0))
                uvlayer[i * 8 + 1].uv = Vector((u1, 1 / 3))
                uvlayer[i * 8 + 2].uv = Vector((u1, 2 / 3))
                uvlayer[i * 8 + 3].uv = Vector((.5, 1))

                uvlayer[i * 8 + 4].uv = Vector((.5, 0))
                uvlayer[i * 8 + 5].uv = Vector((.5, 1))
                uvlayer[i * 8 + 6].uv = Vector((u2, 2 / 3))
                uvlayer[i * 8 + 7].uv = Vector((u2, 1 / 3))

        leafMesh.validate()

    leafVertSize = {'hex': 6, 'rect': 4, 'dFace': 4, 'dVert': 1}[leafShape]

    armLevels = min(armLevels, levels)
    armLevels -= 1

    # unpack vars from splineToBone
    splineToBone1 = splineToBone
    splineToBone = [s[0] if len(s) > 1 else s for s in splineToBone1]
    isend = [s[1] if len(s) > 1 else False for s in splineToBone1]
    issplit = [s[2] if len(s) > 2 else False for s in splineToBone1]
    splitPidx = [s[3] if len(s) > 2 else 0 for s in splineToBone1]

    # If we need an armature we add it
    if useArm:
        # Create the armature and objects
        create_armature(
                    armAnim, leafP, cu, frameRate, leafMesh, leafObj, leafVertSize,
                    leaves, levelCount, splineToBone, treeOb, wind, gust, gustF, af1,
                    af2, af3, leafAnim, loopFrames, previewArm, armLevels, makeMesh, boneStep
                    )

    # print(time.time()-startTime)

    # mesh branches
    if makeMesh:
        t1 = time.time()

        treeMesh = bpy.data.meshes.new('treemesh')
        treeObj = bpy.data.objects.new('treemesh', treeMesh)
        bpy.context.scene.collection.objects.link(treeObj)

        treeVerts = []
        treeEdges = []
        root_vert = []
        vert_radius = []
        vertexGroups = OrderedDict()
        lastVerts = []

        for i, curve in enumerate(cu.splines):
            points = curve.bezier_points

            # find branching level
            level = 0
            for l, c in enumerate(levelCount):
                if i < c:
                    level = l
                    break
            level = min(level, 3)

            step = boneStep[level]
            vindex = len(treeVerts)

            p1 = points[0]

            # add extra vertex for splits
            if issplit[i]:
                pb = int(splineToBone[i][4:-4])
                pn = splitPidx[i]  # int(splineToBone[i][-3:])
                p_1 = cu.splines[pb].bezier_points[pn]
                p_2 = cu.splines[pb].bezier_points[pn + 1]
                p = evalBez(p_1.co, p_1.handle_right, p_2.handle_left, p_2.co, 1 - 1 / (resU + 1))
                treeVerts.append(p)

                root_vert.append(False)
                vert_radius.append((p1.radius * .75, p1.radius * .75))
                treeEdges.append([vindex, vindex + 1])
                vindex += 1

            if isend[i]:
                parent = lastVerts[int(splineToBone[i][4:-4])]
                vindex -= 1
            else:
                # add first point
                treeVerts.append(p1.co)
                root_vert.append(True)
                vert_radius.append((p1.radius, p1.radius))
            """
            # add extra vertex for splits
            if issplit[i]:
                p2 = points[1]
                p = evalBez(p1.co, p1.handle_right, p2.handle_left, p2.co, .001)
                treeVerts.append(p)
                root_vert.append(False)
                vert_radius.append((p1.radius, p1.radius)) #(p1.radius * .95, p1.radius * .95)
                treeEdges.append([vindex,vindex+1])
                vindex += 1
            """
            # dont make vertex group if above armLevels
            if (i >= levelCount[armLevels]):
                idx = i
                groupName = splineToBone[idx]
                g = True
                while groupName not in vertexGroups:
                    # find parent bone of parent bone
                    b = splineToBone[idx]
                    idx = int(b[4:-4])
                    groupName = splineToBone[idx]
            else:
                g = False

            for n, p2 in enumerate(points[1:]):
                if not g:
                    groupName = 'bone' + (str(i)).rjust(3, '0') + '.' + (str(n)).rjust(3, '0')
                    groupName = roundBone(groupName, step)
                    if groupName not in vertexGroups:
                        vertexGroups[groupName] = []

                # parent first vert in split to parent branch bone
                if issplit[i] and n == 0:
                    if g:
                        vertexGroups[groupName].append(vindex - 1)
                    else:
                        vertexGroups[splineToBone[i]].append(vindex - 1)

                for f in range(1, resU + 1):
                    pos = f / resU
                    p = evalBez(p1.co, p1.handle_right, p2.handle_left, p2.co, pos)
                    radius = p1.radius + (p2.radius - p1.radius) * pos

                    treeVerts.append(p)
                    root_vert.append(False)
                    vert_radius.append((radius, radius))

                    if (isend[i]) and (n == 0) and (f == 1):
                        edge = [parent, n * resU + f + vindex]
                    else:
                        edge = [n * resU + f + vindex - 1, n * resU + f + vindex]
                        # add vert to group
                        vertexGroups[groupName].append(n * resU + f + vindex - 1)
                    treeEdges.append(edge)

                vertexGroups[groupName].append(n * resU + resU + vindex)

                p1 = p2

            lastVerts.append(len(treeVerts) - 1)

        treeMesh.from_pydata(treeVerts, treeEdges, ())

        for group in vertexGroups:
            treeObj.vertex_groups.new(name=group)
            treeObj.vertex_groups[group].add(vertexGroups[group], 1.0, 'ADD')

        # add armature
        if useArm:
            armMod = treeObj.modifiers.new('windSway', 'ARMATURE')
            if previewArm:
                bpy.data.objects['treeArm'].hide_viewport = True
                bpy.data.armatures['tree'].display_type = 'STICK'
            armMod.object = bpy.data.objects['treeArm']
            armMod.use_bone_envelopes = False
            armMod.use_vertex_groups = True
            treeObj.parent = bpy.data.objects['treeArm']

        # add skin modifier and set data
        skinMod = treeObj.modifiers.new('Skin', 'SKIN')
        skinMod.use_smooth_shade = True
        if previewArm:
            skinMod.show_viewport = False
        skindata = treeObj.data.skin_vertices[0].data
        for i, radius in enumerate(vert_radius):
            skindata[i].radius = radius
            skindata[i].use_root = root_vert[i]

        print("mesh time", time.time() - t1)



# ##### BEGIN GPL LICENSE BLOCK #####

#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>
import bpy
from bpy.types import Menu, Panel, UIList
from bl_ui.utils import PresetPanel

from bpy.app.translations import pgettext_tip as tip_


class RENDER_PT_presets(PresetPanel, Panel):
    bl_label = "Render Presets"
    preset_subdir = "render"
    preset_operator = "script.execute_preset"
    preset_add_operator = "render.preset_add"


class RENDER_PT_ffmpeg_presets(PresetPanel, Panel):
    bl_label = "FFMPEG Presets"
    preset_subdir = "ffmpeg"
    preset_operator = "script.python_file_run"


class RENDER_MT_framerate_presets(Menu):
    bl_label = "Frame Rate Presets"
    preset_subdir = "framerate"
    preset_operator = "script.execute_preset"
    draw = Menu.draw_preset


class RenderOutputButtonsPanel:
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "output"
    # COMPAT_ENGINES must be defined in each subclass, external engines can add themselves here

    @classmethod
    def poll(cls, context):
        return (context.engine in cls.COMPAT_ENGINES)


class RENDER_PT_dimensions(RenderOutputButtonsPanel, Panel):
    bl_label = "Dimensions"
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    _frame_rate_args_prev = None
    _preset_class = None

    def draw_header_preset(self, _context):
        RENDER_PT_presets.draw_panel_header(self.layout)

    @staticmethod
    def _draw_framerate_label(*args):
        # avoids re-creating text string each draw
        if RENDER_PT_dimensions._frame_rate_args_prev == args:
            return RENDER_PT_dimensions._frame_rate_ret

        fps, fps_base, preset_label = args

        if fps_base == 1.0:
            fps_rate = round(fps)
        else:
            fps_rate = round(fps / fps_base, 2)

        # TODO: Change the following to iterate over existing presets
        custom_framerate = (fps_rate not in {23.98, 24, 25, 29.97, 30, 50, 59.94, 60, 120, 240})

        if custom_framerate is True:
            fps_label_text = tip_("Custom (%.4g fps)") % fps_rate
            show_framerate = True
        else:
            fps_label_text = tip_("%.4g fps") % fps_rate
            show_framerate = (preset_label == "Custom")

        RENDER_PT_dimensions._frame_rate_args_prev = args
        RENDER_PT_dimensions._frame_rate_ret = args = (fps_label_text, show_framerate)
        return args

    @staticmethod
    def draw_framerate(layout, rd):
        if RENDER_PT_dimensions._preset_class is None:
            RENDER_PT_dimensions._preset_class = bpy.types.RENDER_MT_framerate_presets

        args = rd.fps, rd.fps_base, RENDER_PT_dimensions._preset_class.bl_label
        fps_label_text, show_framerate = RENDER_PT_dimensions._draw_framerate_label(*args)

        layout.menu("RENDER_MT_framerate_presets", text=fps_label_text)

        if show_framerate:
            col = layout.column(align=True)
            col.prop(rd, "fps")
            col.prop(rd, "fps_base", text="Base")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False  # No animation.

        scene = context.scene
        rd = scene.render

        col = layout.column(align=True)
        col.prop(rd, "resolution_x", text="Resolution X")
        col.prop(rd, "resolution_y", text="Y")
        col.prop(rd, "resolution_percentage", text="%")

        col = layout.column(align=True)
        col.prop(rd, "pixel_aspect_x", text="Aspect X")
        col.prop(rd, "pixel_aspect_y", text="Y")

        col = layout.column(align=True)
        col.prop(rd, "use_border")
        sub = col.column(align=True)
        sub.active = rd.use_border
        sub.prop(rd, "use_crop_to_border")

        col = layout.column(align=True)
        col.prop(scene, "frame_start", text="Frame Start")
        col.prop(scene, "frame_end", text="End")
        col.prop(scene, "frame_step", text="Step")

        col = layout.column(heading="Frame Rate")
        self.draw_framerate(col, rd)


class RENDER_PT_frame_remapping(RenderOutputButtonsPanel, Panel):
    bl_label = "Time Remapping"
    bl_parent_id = "RENDER_PT_dimensions"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False  # No animation.

        rd = context.scene.render

        col = layout.column(align=True)
        col.prop(rd, "frame_map_old", text="Old")
        col.prop(rd, "frame_map_new", text="New")


class RENDER_PT_post_processing(RenderOutputButtonsPanel, Panel):
    bl_label = "Post Processing"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        rd = context.scene.render

        col = layout.column(heading="Pipeline")
        col.prop(rd, "use_compositing")
        col.prop(rd, "use_sequencer")

        layout.prop(rd, "dither_intensity", text="Dither", slider=True)


class RENDER_PT_stamp(RenderOutputButtonsPanel, Panel):
    bl_label = "Metadata"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False  # No animation.

        rd = context.scene.render

        if rd.use_sequencer:
            layout.prop(rd, "metadata_input")

        col = layout.column(heading="Include")
        col.prop(rd, "use_stamp_date", text="Date")
        col.prop(rd, "use_stamp_time", text="Time")
        col.prop(rd, "use_stamp_render_time", text="Render Time")
        col.prop(rd, "use_stamp_frame", text="Frame")
        col.prop(rd, "use_stamp_frame_range", text="Frame Range")
        col.prop(rd, "use_stamp_memory", text="Memory")
        col.prop(rd, "use_stamp_hostname", text="Hostname")
        col.prop(rd, "use_stamp_camera", text="Camera")
        col.prop(rd, "use_stamp_lens", text="Lens")
        col.prop(rd, "use_stamp_scene", text="Scene")
        col.prop(rd, "use_stamp_marker", text="Marker")
        col.prop(rd, "use_stamp_filename", text="Filename")

        if rd.use_sequencer:
            col.prop(rd, "use_stamp_sequencer_strip", text="Strip Name")


class RENDER_PT_stamp_note(RenderOutputButtonsPanel, Panel):
    bl_label = "Note"
    bl_parent_id = "RENDER_PT_stamp"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    def draw_header(self, context):
        rd = context.scene.render

        self.layout.prop(rd, "use_stamp_note", text="")

    def draw(self, context):
        layout = self.layout

        rd = context.scene.render

        layout.active = rd.use_stamp_note
        layout.prop(rd, "stamp_note_text", text="")


class RENDER_PT_stamp_burn(RenderOutputButtonsPanel, Panel):
    bl_label = "Burn Into Image"
    bl_parent_id = "RENDER_PT_stamp"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    def draw_header(self, context):
        rd = context.scene.render

        self.layout.prop(rd, "use_stamp", text="")

    def draw(self, context):
        layout = self.layout

        rd = context.scene.render

        layout.use_property_split = True

        col = layout.column()
        col.active = rd.use_stamp
        col.prop(rd, "stamp_font_size", text="Font Size")
        col.column().prop(rd, "stamp_foreground", slider=True)
        col.column().prop(rd, "stamp_background", slider=True)
        col.prop(rd, "use_stamp_labels", text="Include Labels")


class RENDER_PT_output(RenderOutputButtonsPanel, Panel):
    bl_label = "Output"
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = False
        layout.use_property_decorate = False  # No animation.

        rd = context.scene.render
        image_settings = rd.image_settings

        layout.prop(rd, "filepath", text="")

        layout.use_property_split = True

        col = layout.column(heading="Saving")
        col.prop(rd, "use_file_extension")
        col.prop(rd, "use_render_cache")

        layout.template_image_settings(image_settings, color_management=False)

        if not rd.is_movie_format:
            col = layout.column(heading="Image Sequence")
            col.prop(rd, "use_overwrite")
            col.prop(rd, "use_placeholder")


class RENDER_PT_output_views(RenderOutputButtonsPanel, Panel):
    bl_label = "Views"
    bl_parent_id = "RENDER_PT_output"
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    @classmethod
    def poll(cls, context):
        rd = context.scene.render
        return rd.use_multiview

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = False
        layout.use_property_decorate = False  # No animation.

        rd = context.scene.render
        layout.template_image_views(rd.image_settings)


class RENDER_PT_encoding(RenderOutputButtonsPanel, Panel):
    bl_label = "Encoding"
    bl_parent_id = "RENDER_PT_output"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    def draw_header_preset(self, _context):
        RENDER_PT_ffmpeg_presets.draw_panel_header(self.layout)

    @classmethod
    def poll(cls, context):
        rd = context.scene.render
        return rd.image_settings.file_format in {'FFMPEG', 'XVID', 'H264', 'THEORA'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        rd = context.scene.render
        ffmpeg = rd.ffmpeg

        layout.prop(rd.ffmpeg, "format")
        layout.prop(ffmpeg, "use_autosplit")


class RENDER_PT_encoding_video(RenderOutputButtonsPanel, Panel):
    bl_label = "Video"
    bl_parent_id = "RENDER_PT_encoding"
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    @classmethod
    def poll(cls, context):
        rd = context.scene.render
        return rd.image_settings.file_format in {'FFMPEG', 'XVID', 'H264', 'THEORA'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        self.draw_vcodec(context)

    def draw_vcodec(self, context):
        """Video codec options."""
        layout = self.layout
        ffmpeg = context.scene.render.ffmpeg

        needs_codec = ffmpeg.format in {'AVI', 'QUICKTIME', 'MKV', 'OGG', 'MPEG4', 'WEBM'}
        if needs_codec:
            layout.prop(ffmpeg, "codec")

        if needs_codec and ffmpeg.codec == 'NONE':
            return

        if ffmpeg.codec == 'DNXHD':
            layout.prop(ffmpeg, "use_lossless_output")

        # Output quality
        use_crf = needs_codec and ffmpeg.codec in {'H264', 'MPEG4', 'WEBM'}
        if use_crf:
            layout.prop(ffmpeg, "constant_rate_factor")

        # Encoding speed
        layout.prop(ffmpeg, "ffmpeg_preset")
        # I-frames
        layout.prop(ffmpeg, "gopsize")
        # B-Frames
        row = layout.row(align=True, heading="Max B-frames")
        row.prop(ffmpeg, "use_max_b_frames", text="")
        sub = row.row(align=True)
        sub.active = ffmpeg.use_max_b_frames
        sub.prop(ffmpeg, "max_b_frames", text="")

        if not use_crf or ffmpeg.constant_rate_factor == 'NONE':
            col = layout.column()

            sub = col.column(align=True)
            sub.prop(ffmpeg, "video_bitrate")
            sub.prop(ffmpeg, "minrate", text="Minimum")
            sub.prop(ffmpeg, "maxrate", text="Maximum")

            col.prop(ffmpeg, "buffersize", text="Buffer")

            col.separator()

            col.prop(ffmpeg, "muxrate", text="Mux Rate")
            col.prop(ffmpeg, "packetsize", text="Mux Packet Size")


class RENDER_PT_encoding_audio(RenderOutputButtonsPanel, Panel):
    bl_label = "Audio"
    bl_parent_id = "RENDER_PT_encoding"
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}

    @classmethod
    def poll(cls, context):
        rd = context.scene.render
        return rd.image_settings.file_format in {'FFMPEG', 'XVID', 'H264', 'THEORA'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        rd = context.scene.render
        ffmpeg = rd.ffmpeg

        if ffmpeg.format != 'MP3':
            layout.prop(ffmpeg, "audio_codec", text="Audio Codec")

        if ffmpeg.audio_codec != 'NONE':
            layout.prop(ffmpeg, "audio_channels")
            layout.prop(ffmpeg, "audio_mixrate", text="Sample Rate")
            layout.prop(ffmpeg, "audio_bitrate")
            layout.prop(ffmpeg, "audio_volume", slider=True)


class RENDER_UL_renderviews(UIList):
    def draw_item(self, _context, layout, _data, item, icon, _active_data, _active_propname, index):
        view = item
        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            if view.name in {"left", "right"}:
                layout.label(text=view.name, icon_value=icon + (not view.use))
            else:
                layout.prop(view, "name", text="", index=index, icon_value=icon, emboss=False)
            layout.prop(view, "use", text="", index=index)

        elif self.layout_type == 'GRID':
            layout.alignment = 'CENTER'
            layout.label(text="", icon_value=icon + (not view.use))


class RENDER_PT_stereoscopy(RenderOutputButtonsPanel, Panel):
    bl_label = "Stereoscopy"
    COMPAT_ENGINES = {'BLENDER_RENDER', 'BLENDER_EEVEE', 'BLENDER_WORKBENCH'}
    bl_options = {'DEFAULT_CLOSED'}

    def draw_header(self, context):
        rd = context.scene.render
        self.layout.prop(rd, "use_multiview", text="")

    def draw(self, context):
        layout = self.layout

        scene = context.scene
        rd = scene.render
        rv = rd.views.active

        layout.active = rd.use_multiview
        basic_stereo = rd.views_format == 'STEREO_3D'

        row = layout.row()
        layout.row().prop(rd, "views_format", expand=True)

        if basic_stereo:
            row = layout.row()
            row.template_list("RENDER_UL_renderviews", "name", rd, "stereo_views", rd.views, "active_index", rows=2)

            row = layout.row()
            row.use_property_split = True
            row.use_property_decorate = False
            row.prop(rv, "file_suffix")

        else:
            row = layout.row()
            row.template_list("RENDER_UL_renderviews", "name", rd, "views", rd.views, "active_index", rows=2)

            col = row.column(align=True)
            col.operator("scene.render_view_add", icon='ADD', text="")
            col.operator("scene.render_view_remove", icon='REMOVE', text="")

            row = layout.row()
            row.use_property_split = True
            row.use_property_decorate = False
            row.prop(rv, "camera_suffix")


classes = (
    RENDER_PT_presets,
    RENDER_PT_ffmpeg_presets,
    RENDER_MT_framerate_presets,
    RENDER_PT_dimensions,
    RENDER_PT_frame_remapping,
    RENDER_PT_stereoscopy,
    RENDER_PT_output,
    RENDER_PT_output_views,
    RENDER_PT_encoding,
    RENDER_PT_encoding_video,
    RENDER_PT_encoding_audio,
    RENDER_PT_stamp,
    RENDER_PT_stamp_note,
    RENDER_PT_stamp_burn,
    RENDER_UL_renderviews,
    RENDER_PT_post_processing,
)

if __name__ == "__main__":  # only for live edit.
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)



#====================== BEGIN GPL LICENSE BLOCK ======================
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#======================= END GPL LICENSE BLOCK ========================

# <pep8 compliant>

import collections

from types import FunctionType
from itertools import chain


#=============================================
# Class With Stages
#=============================================


def rigify_stage(stage):
    """Decorates the method with the specified stage."""
    def process(method):
        if not isinstance(method, FunctionType):
            raise ValueError("Stage decorator must be applied to a method definition")
        method._rigify_stage = stage
        return method
    return process


class StagedMetaclass(type):
    """
    Metaclass for rigs that manages assignment of methods to stages via @stage.* decorators.

    Using define_stages=True in the class definition will register all non-system
    method names from that definition as valid stages. After that, subclasses can
    register methods to those stages, to be called via rigify_invoke_stage.
    """
    def __new__(metacls, class_name, bases, namespace, define_stages=None, **kwds):
        # suppress keyword args to avoid issues with __init_subclass__
        return super().__new__(metacls, class_name, bases, namespace, **kwds)

    def __init__(self, class_name, bases, namespace, define_stages=None, **kwds):
        super().__init__(class_name, bases, namespace, **kwds)

        # Compute the set of stages defined by this class
        if not define_stages:
            define_stages = []

        elif define_stages is True:
            define_stages = [
                name for name, item in namespace.items()
                if name[0] != '_' and isinstance(item, FunctionType)
            ]

        self.rigify_own_stages = frozenset(define_stages)

        # Compute complete set of inherited stages
        staged_bases = [ cls for cls in reversed(self.__mro__) if isinstance(cls, StagedMetaclass) ]

        self.rigify_stages = stages = frozenset(chain.from_iterable(
            cls.rigify_own_stages for cls in staged_bases
        ))

        # Compute the inherited stage to method mapping
        stage_map = collections.defaultdict(collections.OrderedDict)
        own_stage_map = collections.defaultdict(collections.OrderedDict)
        method_map = {}

        self.rigify_own_stage_map = own_stage_map

        for base in staged_bases:
            for stage_name, methods in base.rigify_own_stage_map.items():
                for method_name, method_class in methods.items():
                    if method_name in stages:
                        raise ValueError("Stage method '%s' inherited @stage.%s in class %s (%s)" %
                                         (method_name, stage_name, class_name, self.__module__))

                    # Check consistency of inherited stage assignment to methods
                    if method_name in method_map:
                        if method_map[method_name] != stage_name:
                            print("RIGIFY CLASS %s (%s): method '%s' has inherited both @stage.%s and @stage.%s\n" %
                                  (class_name, self.__module__, method_name, method_map[method_name], stage_name))
                    else:
                        method_map[method_name] = stage_name

                    stage_map[stage_name][method_name] = method_class

        # Scan newly defined methods for stage decorations
        for method_name, item in namespace.items():
            if isinstance(item, FunctionType):
                stage = getattr(item, '_rigify_stage', None)

                if stage and method_name in stages:
                    print("RIGIFY CLASS %s (%s): cannot use stage decorator on the stage method '%s' (@stage.%s ignored)" %
                            (class_name, self.__module__, method_name, stage))
                    continue

                # Ensure that decorators aren't lost when redefining methods
                if method_name in method_map:
                    if not stage:
                        stage = method_map[method_name]
                        print("RIGIFY CLASS %s (%s): missing stage decorator on method '%s' (should be @stage.%s)" %
                              (class_name, self.__module__, method_name, stage))
                    # Check that the method is assigned to only one stage
                    elif stage != method_map[method_name]:
                        print("RIGIFY CLASS %s (%s): method '%s' has decorator @stage.%s, but inherited base has @stage.%s" %
                              (class_name, self.__module__, method_name, stage, method_map[method_name]))

                # Assign the method to the stage, verifying that it's valid
                if stage:
                    if stage not in stages:
                        raise ValueError("Invalid stage name '%s' for method '%s' in class %s (%s)" %
                                         (stage, method_name, class_name, self.__module__))
                    else:
                        stage_map[stage][method_name] = self
                        own_stage_map[stage][method_name] = self

        self.rigify_stage_map = stage_map

    def make_stage_decorators(self):
        return [(name, rigify_stage(name)) for name in self.rigify_stages]


class BaseStagedClass(object, metaclass=StagedMetaclass):
    rigify_sub_objects = tuple()
    rigify_sub_object_run_late = False

    def rigify_invoke_stage(self, stage):
        """Call all methods decorated with the given stage, followed by the callback."""
        cls = self.__class__
        assert isinstance(cls, StagedMetaclass)
        assert stage in cls.rigify_stages

        getattr(self, stage)()

        for sub in self.rigify_sub_objects:
            if not sub.rigify_sub_object_run_late:
                sub.rigify_invoke_stage(stage)

        for method_name in cls.rigify_stage_map[stage]:
            getattr(self, method_name)()

        for sub in self.rigify_sub_objects:
            if sub.rigify_sub_object_run_late:
                sub.rigify_invoke_stage(stage)


#=============================================
# Per-owner singleton class
#=============================================


class SingletonPluginMetaclass(StagedMetaclass):
    """Metaclass for maintaining one instance per owner object per constructor arg set."""
    def __call__(cls, owner, *constructor_args):
        key = (cls, *constructor_args)
        try:
            return owner.plugin_map[key]
        except KeyError:
            new_obj = super().__call__(owner, *constructor_args)
            owner.plugin_map[key] = new_obj
            owner.plugin_list.append(new_obj)
            owner.plugin_list.sort(key=lambda obj: obj.priority, reverse=True)
            return new_obj



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

# To run the test type: blender -b /path/to/the/blend/file --python path/to/this/py/file -- --run-all-tests -- --verbose
# Type the above line in cmd/terminal, for example, look below
# blender -b c:\blender-lib\deform_modifiers.blend --python c:\deform_modifiers.py -- --run-all-tests -- --verbose


import os
import sys
import bpy

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from modules.mesh_test import MeshTest, ModifierSpec, OperatorSpecObjectMode, DeformModifierSpec, RunTest

tests = [

    # Surface Deform Test, finally can bind to the Target object.
    # Actual deformation occurs by animating imitating user input.

    MeshTest("SurfaceDeform", "testObjMonkeySurfaceDeform", "expObjMonkeySurfaceDeform",
             [DeformModifierSpec(10, [
                 ModifierSpec('surface_deform', 'SURFACE_DEFORM', {'target': bpy.data.objects["Cube"]})],
                                 OperatorSpecObjectMode('surfacedeform_bind', {'modifier': 'surface_deform'}))]),

    # Mesh Deform Test, finally can bind to the Target object.
    # Actual deformation occurs by animating imitating user input.

    MeshTest("MeshDeform", "testObjMonkeyMeshDeform", "expObjMonkeyMeshDeform",
             [DeformModifierSpec(10, [ModifierSpec('mesh_deform', 'MESH_DEFORM',
                                                   {'object': bpy.data.objects["MeshCube"], 'precision': 2})],
                                 OperatorSpecObjectMode('meshdeform_bind', {'modifier': 'mesh_deform'}))]),

    # Surface Deform Test, finally can bind to the Target object.
    # Actual deformation occurs by animating imitating user input.

    MeshTest("Hook", "testObjHookPlane", "expObjHookPlane",
             [DeformModifierSpec(10, [ModifierSpec('hook', 'HOOK',
                                                   {'object': bpy.data.objects["Empty"], 'falloff_radius': 1,
                                                    'vertex_group': 'Group'})])]),

    # Laplacian Deform Test, first a hook is attached.

    MeshTest("Laplace", "testObjCubeLaplacian", "expObjCubeLaplacian",
             [DeformModifierSpec(10,
                                 [ModifierSpec('hook2', 'HOOK', {'object': bpy.data.objects["Empty.001"],
                                                                 'vertex_group': 'hook_vg'}),
                                  ModifierSpec('laplace', 'LAPLACIANDEFORM', {'vertex_group': 'laplace_vg'})],
                                 OperatorSpecObjectMode('laplaciandeform_bind', {'modifier': 'laplace'}))]),

    MeshTest("WarpPlane", "testObjPlaneWarp", "expObjPlaneWarp",
             [DeformModifierSpec(10, [ModifierSpec('warp', 'WARP',
                                                   {'object_from': bpy.data.objects["From"],
                                                    'object_to': bpy.data.objects["To"],
                                                    })])]),

    #############################################
    # Curves Deform Modifiers
    #############################################
    MeshTest("CurveArmature", "testObjBezierCurveArmature", "expObjBezierCurveArmature",
             [DeformModifierSpec(10, [ModifierSpec('curve_armature', 'ARMATURE',
                                                   {'object': bpy.data.objects['testArmatureHelper'],
                                                    'use_vertex_groups': False, 'use_bone_envelopes': True})])]),

    MeshTest("CurveLattice", "testObjBezierCurveLattice", "expObjBezierCurveLattice",
             [DeformModifierSpec(10, [ModifierSpec('curve_lattice', 'LATTICE',
                                                   {'object': bpy.data.objects['testLatticeCurve']})])]),

    # HOOK for Curves can't be tested with current framework, as it requires going to Edit Mode to select vertices,
    # here is no equivalent of a vertex group in Curves.
    # Dummy test for Hook, can also be called corner case
    MeshTest("CurveHook", "testObjBezierCurveHook", "expObjBezierCurveHook",
             [DeformModifierSpec(10,
                                 [ModifierSpec('curve_Hook', 'HOOK', {'object': bpy.data.objects['EmptyCurve']})])]),

    MeshTest("MeshDeformCurve", "testObjCurveMeshDeform", "expObjCurveMeshDeform",
             [DeformModifierSpec(10, [
                 ModifierSpec('mesh_deform_curve', 'MESH_DEFORM', {'object': bpy.data.objects["Cylinder"],
                                                                   'precision': 2})],
                                 OperatorSpecObjectMode('meshdeform_bind', {'modifier': 'mesh_deform_curve'}))]),

    MeshTest("WarpCurve", "testObjBezierCurveWarp", "expObjBezierCurveWarp",
             [DeformModifierSpec(10, [ModifierSpec('warp_curve', 'WARP',
                                                   {'object_from': bpy.data.objects["From_curve"],
                                                    'object_to': bpy.data.objects["To_curve"]})])]),

]

deform_tests = RunTest(tests)
command = list(sys.argv)
for i, cmd in enumerate(command):
    if cmd == "--run-all-tests":
        deform_tests.apply_modifiers = True
        deform_tests.do_compare = True
        deform_tests.run_all_tests()
        break
    elif cmd == "--run-test":
        deform_tests.apply_modifiers = False
        deform_tests.do_compare = False
        name = command[i + 1]
        deform_tests.run_test(name)
        break
